# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module contains error classes required by IQMClient.
"""


class ClientAuthenticationError(RuntimeError):
    """Something went wrong with user authentication."""


class ClientConfigurationError(RuntimeError):
    """Wrong configuration provided."""


class CircuitValidationError(RuntimeError):
    """Circuit validation failed."""


class CircuitTranspilationError(RuntimeError):
    """Circuit transpilation failed."""


class CircuitExecutionError(RuntimeError):
    """Something went wrong on the server."""


class APITimeoutError(CircuitExecutionError):
    """Executing a job on the server took too long."""


class JobAbortionError(RuntimeError):
    """Job abortion failed."""


class ArchitectureRetrievalError(RuntimeError):
    """Retrieving quantum architecture failed."""



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the IQM client.
"""
# pylint: disable=too-many-arguments,too-many-lines
from importlib.metadata import version
import re
import uuid

from mockito import ANY, expect, unstub, verifyNoUnwantedInteractions, when
from packaging.version import parse
import pytest
import requests
from requests import HTTPError

from iqm.iqm_client import (
    APIEndpoint,
    ArchitectureRetrievalError,
    Circuit,
    CircuitCompilationOptions,
    CircuitExecutionError,
    CircuitValidationError,
    ClientConfigurationError,
    Counts,
    DynamicQuantumArchitecture,
    HeraldingMode,
    Instruction,
    IQMClient,
    JobAbortionError,
    QuantumArchitectureSpecification,
    SingleQubitMapping,
    Status,
    serialize_qubit_mapping,
    validate_circuit,
)
from tests.conftest import (
    MockJsonResponse,
    get_jobs_args,
    mock_supported_client_libraries_response,
    post_jobs_args,
    submit_circuits_args,
)


@pytest.fixture
def move_circuit():
    instructions = (
        Instruction(
            name='prx',
            qubits=('QB1',),
            args={'phase_t': 0.3, 'angle_t': -0.2},
        ),
        Instruction(
            name='move',
            qubits=('QB3', 'COMP_R'),
            args={},
        ),
        Instruction(
            name='cz',
            qubits=('QB1', 'COMP_R'),
            args={},
        ),
        Instruction(
            name='cz',
            qubits=('QB2', 'COMP_R'),
            args={},
        ),
        Instruction(
            name='move',
            qubits=('QB3', 'COMP_R'),
            args={},
        ),
    )
    return Circuit(name='COMP_R circuit', instructions=instructions)


@pytest.fixture
def move_circuit_with_prx_in_the_sandwich():
    instructions = (
        Instruction(
            name='prx',
            qubits=('QB1',),
            args={'phase_t': 0.3, 'angle_t': -0.2},
        ),
        Instruction(
            name='move',
            qubits=('QB3', 'COMP_R'),
            args={},
        ),
        Instruction(
            name='prx',
            qubits=('QB3',),
            args={'phase_t': 0.3, 'angle_t': -0.2},
        ),
        Instruction(
            name='move',
            qubits=('QB3', 'COMP_R'),
            args={},
        ),
    )
    return Circuit(name='COMP_R circuit with PRX in the sandwich', instructions=instructions)


def test_serialize_qubit_mapping():
    """
    Tests that serialize_qubit_mapping returns a list of SingleQubitMapping objects
    """
    qm_dict = {'Alice': 'QB1', 'Bob': 'qubit_3', 'Charlie': 'physical 0'}
    assert serialize_qubit_mapping(qm_dict) == [
        SingleQubitMapping(logical_name='Alice', physical_name='QB1'),
        SingleQubitMapping(logical_name='Bob', physical_name='qubit_3'),
        SingleQubitMapping(logical_name='Charlie', physical_name='physical 0'),
    ]


def test_submit_circuits_adds_user_agent(
    sample_client, jobs_url, minimal_run_request, submit_success, dynamic_architecture_url, dynamic_architecture_success
):
    """
    Tests that submit_circuit without client signature adds correct User-Agent header
    """
    expect(requests, times=1).post(jobs_url, **post_jobs_args(minimal_run_request)).thenReturn(submit_success)
    expect(requests, times=1).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    sample_client.submit_circuits(**submit_circuits_args(minimal_run_request))

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_adds_user_agent_with_client_signature(
    client_with_signature,
    jobs_url,
    minimal_run_request,
    submit_success,
    dynamic_architecture_url,
    dynamic_architecture_success,
):
    """
    Tests that submit_circuit with client signature adds correct User-Agent header
    """
    expect(requests, times=1).post(
        jobs_url, **post_jobs_args(minimal_run_request, user_agent=client_with_signature._signature)
    ).thenReturn(submit_success)
    expect(requests, times=1).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    client_with_signature.submit_circuits(**submit_circuits_args(minimal_run_request))

    verifyNoUnwantedInteractions()
    unstub()


@pytest.mark.parametrize(
    'run_request_name, valid_request, error',
    [
        ('minimal_run_request', True, None),
        ('run_request_with_heralding', True, None),
        ('run_request_with_custom_settings', True, None),
        (
            'run_request_with_invalid_qubit_mapping',
            False,
            CircuitValidationError('Multiple logical qubits map to the same physical qubit.'),
        ),
        (
            'run_request_with_incomplete_qubit_mapping',
            False,
            CircuitValidationError(
                "The qubits {'Qubit B'} in circuit 'The circuit' "
                'at index 0 are not found in the provided qubit mapping.'
            ),
        ),
        ('run_request_without_qubit_mapping', True, None),
        ('run_request_with_calibration_set_id', True, None),
        ('run_request_with_duration_check_disabled', True, None),
        (
            'run_request_with_incompatible_options',
            False,
            ValueError(
                'Unable to perform full MOVE gate frame tracking if MOVE gate validation '
                'is not "strict" or "allow_prx".'
            ),
        ),
    ],
)
def test_submit_circuits_returns_id(
    sample_client,
    jobs_url,
    run_request_name,
    valid_request,
    error,
    request,
    submit_success,
    existing_run_id,
    dynamic_architecture_success,
):
    """
    Tests submitting circuits for execution
    """
    run_request = request.getfixturevalue(run_request_name)
    calibration_set_id = uuid.UUID('ec6f6478-a99b-4e75-8a94-2f9cb0511bce')
    run_request.calibration_set_id = calibration_set_id
    dynamic_architecture_success.json_data['calibration_set_id'] = calibration_set_id
    when(requests).get(sample_client._api.url(APIEndpoint.CALIBRATED_GATES, str(calibration_set_id)), ...).thenReturn(
        dynamic_architecture_success
    )

    if valid_request:
        expect(requests, times=1).post(jobs_url, **post_jobs_args(run_request)).thenReturn(submit_success)
    if error is None:
        assert sample_client.submit_circuits(**submit_circuits_args(run_request)) == existing_run_id
    else:
        with pytest.raises(type(error), match=str(error)):
            sample_client.submit_circuits(**submit_circuits_args(run_request))

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_does_not_activate_heralding_by_default(
    sample_client, jobs_url, minimal_run_request, submit_success, dynamic_architecture_url, dynamic_architecture_success
):
    """
    Test submitting run request without heralding
    """
    # Expect request to have heralding mode NONE by default
    assert post_jobs_args(minimal_run_request)['json']['heralding_mode'] == HeraldingMode.NONE.value
    expect(requests, times=1).post(jobs_url, **post_jobs_args(minimal_run_request)).thenReturn(submit_success)
    when(requests).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    # Specify no heralding mode in submit_circuits
    sample_client.submit_circuits(circuits=minimal_run_request.circuits, shots=minimal_run_request.shots)

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_raises_with_invalid_shots(sample_client, minimal_run_request):
    """
    Test that submitting run request with invalid number of shots raises ValueError
    """
    args = submit_circuits_args(minimal_run_request)
    args['shots'] = 0
    with pytest.raises(ValueError, match='Number of shots must be greater than zero.'):
        sample_client.submit_circuits(**args)


def test_submit_circuits_sets_heralding_mode_in_run_request(
    sample_client,
    jobs_url,
    run_request_with_heralding,
    submit_success,
    dynamic_architecture_url,
    dynamic_architecture_success,
):
    """
    Test submitting run request with heralding
    """
    # Expect heralding mode to be the same as in run request
    expected_heralding_mode = run_request_with_heralding.heralding_mode.value
    assert post_jobs_args(run_request_with_heralding)['json']['heralding_mode'] == expected_heralding_mode
    expect(requests, times=1).post(jobs_url, **post_jobs_args(run_request_with_heralding)).thenReturn(submit_success)
    expect(requests, times=1).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    assert submit_circuits_args(run_request_with_heralding)['options'].heralding_mode == expected_heralding_mode
    sample_client.submit_circuits(**submit_circuits_args(run_request_with_heralding))

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_gets_architecture_once(
    sample_client,
    jobs_url,
    minimal_run_request,
    submit_success,
    dynamic_architecture_success,
):
    """
    Test that dynamic quantum architecture is only requested once from the QC when calset id is specified
    """
    calibration_set_id = uuid.UUID('ec6f6478-a99b-4e75-8a94-2f9cb0511bce')
    minimal_run_request.calibration_set_id = calibration_set_id
    dynamic_architecture_success.json_data['calibration_set_id'] = calibration_set_id
    expect(requests, times=1).get(
        sample_client._api.url(APIEndpoint.CALIBRATED_GATES, str(calibration_set_id)), ...
    ).thenReturn(dynamic_architecture_success)
    expect(requests, times=1).post(jobs_url, **post_jobs_args(minimal_run_request)).thenReturn(submit_success)
    # Get architecture explicitly and then submit job
    sample_client.get_dynamic_quantum_architecture(calibration_set_id)
    sample_client.submit_circuits(**submit_circuits_args(minimal_run_request))
    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_raises_with_invalid_heralding_mode(
    sample_client,
    dynamic_architecture_url,
    dynamic_architecture_success,
):
    """
    Test that submitting run request with invalid heralding mode raises an error
    """
    when(requests).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)
    with pytest.raises(ValueError, match="Input should be 'none' or 'zeros'"):
        sample_client.submit_circuits(
            circuits=[], shots=10, options=CircuitCompilationOptions(heralding_mode='invalid')
        )


def test_get_run_adds_user_agent(
    sample_client,
    existing_job_url,
    existing_run_id,
    pending_compilation_job_result,
    dynamic_architecture_url,
    dynamic_architecture_success,
):
    """
    Tests that get_run without client signature adds the correct User-Agent header
    """
    expect(requests, times=1).get(existing_job_url, **get_jobs_args()).thenReturn(pending_compilation_job_result)
    when(requests).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    sample_client.get_run(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_adds_user_agent_with_client_signature(
    client_with_signature, client_signature, existing_job_url, existing_run_id, pending_compilation_job_result
):
    """
    Tests that get_run with client signature adds the correct User-Agent header
    """
    assert client_signature in client_with_signature._signature
    expect(requests, times=1).get(
        existing_job_url,
        **get_jobs_args(user_agent=client_with_signature._signature),
    ).thenReturn(pending_compilation_job_result)

    client_with_signature.get_run(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_status_and_results_for_existing_run(
    sample_client,
    existing_job_url,
    sample_circuit_metadata,
    sample_calibration_set_id,
    pending_compilation_job_result,
    pending_execution_job_result,
    pending_deletion_job_result,
    deleted_job_result,
    ready_job_result,
    existing_run_id,
):
    """
    Tests getting the run status
    """
    expect(requests, times=5).get(existing_job_url, **get_jobs_args()).thenReturn(
        pending_compilation_job_result
    ).thenReturn(pending_execution_job_result).thenReturn(ready_job_result).thenReturn(
        pending_deletion_job_result
    ).thenReturn(
        deleted_job_result
    )

    # First request gets status 'pending compilation'
    assert sample_client.get_run(existing_run_id).status == Status.PENDING_COMPILATION

    # Second requests gets status 'pending execution'
    assert sample_client.get_run(existing_run_id).status == Status.PENDING_EXECUTION

    # Third request gets status 'ready'
    job_result = sample_client.get_run(existing_run_id)
    assert job_result.status == Status.READY
    assert job_result.measurements is not None
    assert job_result.metadata.request.calibration_set_id == sample_calibration_set_id
    assert job_result.metadata.request.circuits[0].metadata == sample_circuit_metadata

    # Fourth request gets status 'pending deletion'
    assert sample_client.get_run(existing_run_id).status == Status.PENDING_DELETION

    # Fifth request gets status 'deleted'
    job_result = sample_client.get_run(existing_run_id)
    assert job_result.status == Status.DELETED
    assert job_result.measurements is None
    assert job_result.metadata.request.circuits == []
    assert job_result.metadata.request.shots == 1

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_status_for_existing_run(
    sample_client,
    existing_job_status_url,
    pending_compilation_status,
    pending_execution_status,
    ready_status,
    pending_deletion_status,
    deleted_status,
    existing_run_id,
):
    """
    Tests getting the run status
    """
    expect(requests, times=5).get(existing_job_status_url, **get_jobs_args()).thenReturn(
        pending_compilation_status
    ).thenReturn(pending_execution_status).thenReturn(ready_status).thenReturn(pending_deletion_status).thenReturn(
        deleted_status
    )

    # First request gets status 'pending compilation'
    assert sample_client.get_run_status(existing_run_id).status == Status.PENDING_COMPILATION
    # Second request gets status 'pending execution'
    assert sample_client.get_run_status(existing_run_id).status == Status.PENDING_EXECUTION
    # Third request gets status 'ready'
    assert sample_client.get_run_status(existing_run_id).status == Status.READY
    # Fourth request gets status 'pending deletion'
    assert sample_client.get_run_status(existing_run_id).status == Status.PENDING_DELETION
    # Fifth request gets status 'deleted'
    assert sample_client.get_run_status(existing_run_id).status == Status.DELETED

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_status_and_results_for_missing_run(sample_client, jobs_url, missing_run_id):
    """
    Tests getting a task that was not created
    """
    expect(requests, times=1).get(f'{jobs_url}/{missing_run_id}', **get_jobs_args()).thenReturn(
        MockJsonResponse(404, {'detail': 'not found'})
    )

    with pytest.raises(HTTPError) as e:
        sample_client.get_run(missing_run_id)
    assert e.value.response.status_code == 404

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_status_for_missing_run(sample_client, jobs_url, missing_run_id):
    """
    Tests getting a task that was not created
    """
    expect(requests, times=1).get(f'{jobs_url}/{missing_run_id}/status', **get_jobs_args()).thenReturn(
        MockJsonResponse(404, {'detail': 'not found'})
    )

    with pytest.raises(HTTPError) as e:
        sample_client.get_run_status(missing_run_id)
    assert e.value.response.status_code == 404

    verifyNoUnwantedInteractions()
    unstub()


def test_waiting_for_compilation(
    sample_client,
    existing_job_status_url,
    pending_compilation_status,
    pending_execution_status,
    existing_job_url,
    pending_execution_job_result,
    existing_run_id,
):
    """
    Tests waiting for compilation for an existing task
    """
    expect(requests, times=3).get(existing_job_status_url, **get_jobs_args()).thenReturn(
        pending_compilation_status
    ).thenReturn(pending_compilation_status).thenReturn(pending_execution_status)
    expect(requests, times=1).get(existing_job_url, **get_jobs_args()).thenReturn(pending_execution_job_result)

    assert sample_client.wait_for_compilation(existing_run_id).status == Status.PENDING_EXECUTION

    verifyNoUnwantedInteractions()
    unstub()


def test_wait_for_compilation_adds_user_agent_with_signature(
    client_with_signature,
    client_signature,
    existing_job_status_url,
    pending_compilation_status,
    pending_execution_status,
    existing_job_url,
    pending_execution_job_result,
    existing_run_id,
):
    """
    Tests that wait_for_compilation with client signature adds the correct User-Agent header
    """
    assert client_signature in client_with_signature._signature
    expect(requests, times=3).get(
        existing_job_status_url, **get_jobs_args(user_agent=client_with_signature._signature)
    ).thenReturn(pending_compilation_status).thenReturn(pending_compilation_status).thenReturn(pending_execution_status)
    expect(requests, times=1).get(
        existing_job_url, **get_jobs_args(user_agent=client_with_signature._signature)
    ).thenReturn(pending_execution_job_result)

    assert client_with_signature.wait_for_compilation(existing_run_id).status == Status.PENDING_EXECUTION

    verifyNoUnwantedInteractions()
    unstub()


def test_waiting_for_results(
    sample_client,
    existing_job_status_url,
    pending_compilation_status,
    pending_execution_status,
    ready_status,
    existing_job_url,
    ready_job_result,
    existing_run_id,
):
    """
    Tests waiting for results for an existing task
    """
    expect(requests, times=3).get(existing_job_status_url, **get_jobs_args()).thenReturn(
        pending_compilation_status
    ).thenReturn(pending_execution_status).thenReturn(ready_status)
    expect(requests, times=1).get(existing_job_url, **get_jobs_args()).thenReturn(ready_job_result)

    assert sample_client.wait_for_results(existing_run_id).status == Status.READY

    verifyNoUnwantedInteractions()
    unstub()


def test_wait_for_results_adds_user_agent_with_signature(
    client_with_signature,
    client_signature,
    existing_job_status_url,
    pending_compilation_status,
    pending_execution_status,
    ready_status,
    existing_job_url,
    ready_job_result,
    existing_run_id,
):
    """
    Tests that wait_for_results without client signature adds the correct User-Agent header
    """
    assert client_signature in client_with_signature._signature
    expect(requests, times=3).get(
        existing_job_status_url, **get_jobs_args(user_agent=client_with_signature._signature)
    ).thenReturn(pending_compilation_status).thenReturn(pending_execution_status).thenReturn(ready_status)
    expect(requests, times=1).get(
        existing_job_url, **get_jobs_args(user_agent=client_with_signature._signature)
    ).thenReturn(ready_job_result)

    assert client_with_signature.wait_for_results(existing_run_id).status == Status.READY

    verifyNoUnwantedInteractions()
    unstub()


def test_get_quantum_architecture(
    sample_client, quantum_architecture_url, sample_static_architecture, static_architecture_success
):
    """Test retrieving the quantum architecture"""
    expect(requests, times=1).get(quantum_architecture_url, ...).thenReturn(static_architecture_success)

    assert sample_client.get_quantum_architecture() == QuantumArchitectureSpecification(
        **sample_static_architecture['quantum_architecture']
    )

    verifyNoUnwantedInteractions()
    unstub()


def test_user_warning_is_emitted_when_warnings_in_response(
    sample_client, existing_job_url, job_result_with_warnings, existing_run_id
):
    """Test that a warning is emitted when warnings are present in the response"""
    expect(requests, times=1).get(existing_job_url, ...).thenReturn(job_result_with_warnings)

    expected_message = job_result_with_warnings.json()['warnings'][0]
    with pytest.warns(UserWarning, match=expected_message):
        sample_client.get_run(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_base_url_is_invalid():
    """Test that an exception is raised when the base URL is invalid"""
    invalid_base_url = 'xyz://example.com'
    with pytest.raises(ClientConfigurationError) as exc:
        IQMClient(invalid_base_url)
    assert f'The URL schema has to be http or https. Incorrect schema in URL: {invalid_base_url}' == str(exc.value)


def test_run_result_throws_json_decode_error_if_received_not_json(
    sample_client, existing_job_url, existing_run_id, not_valid_json_response
):
    """Test that an exception is raised when the response is not a valid JSON"""
    expect(requests, times=1).get(existing_job_url, ...).thenReturn(not_valid_json_response)

    with pytest.raises(CircuitExecutionError):
        sample_client.get_run(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_run_result_status_throws_json_decode_error_if_received_not_json(
    sample_client, existing_job_status_url, existing_run_id, not_valid_json_response
):
    """Test that an exception is raised when the response is not a valid JSON"""
    expect(requests, times=1).get(existing_job_status_url, ...).thenReturn(not_valid_json_response)

    with pytest.raises(CircuitExecutionError):
        sample_client.get_run_status(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_quantum_architecture_throws_json_decode_error_if_received_not_json(
    sample_client, quantum_architecture_url, not_valid_json_response
):
    """Test that an exception is raised when the response is not a valid JSON"""
    expect(requests, times=1).get(quantum_architecture_url, ...).thenReturn(not_valid_json_response)

    with pytest.raises(ArchitectureRetrievalError):
        sample_client.get_quantum_architecture()

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_throws_json_decode_error_if_received_not_json(
    sample_client, jobs_url, not_valid_json_response, dynamic_architecture_url, dynamic_architecture_success
):
    """Test that an exception is raised when the response is not a valid JSON"""
    expect(requests, times=1).post(jobs_url, ...).thenReturn(not_valid_json_response)
    expect(requests, times=1).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    with pytest.raises(CircuitExecutionError):
        sample_client.submit_circuits([])

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_throws_client_configuration_error_on_400(
    sample_client,
    jobs_url,
    not_valid_client_configuration_response,
    dynamic_architecture_url,
    dynamic_architecture_success,
):
    """Test that an exception is raised when the response is 400"""
    expect(requests, times=1).post(jobs_url, ...).thenReturn(not_valid_client_configuration_response)
    expect(requests, times=1).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    with pytest.raises(ClientConfigurationError):
        sample_client.submit_circuits([])

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_validates_circuits(sample_client, sample_circuit):
    """
    Tests that <submit_circuits> validates the batch of provided circuits
    before submitting them for execution
    """
    invalid_circuit = sample_circuit.model_copy()
    invalid_circuit.name = ''  # Invalidate the circuit on purpose
    with pytest.raises(CircuitValidationError, match='The circuit at index 1 failed the validation'):
        sample_client.submit_circuits(circuits=[sample_circuit, invalid_circuit], shots=10)


def test_validate_circuit_accepts_valid_circuit(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    accepts a valid circuit.
    """
    circuit = sample_circuit.model_copy()
    validate_circuit(circuit)


def test_validate_circuit_detects_circuit_name_is_empty_string(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches empty name of a circuit
    """
    circuit = sample_circuit.model_copy()
    circuit.name = ''
    with pytest.raises(ValueError, match='A circuit should have a non-empty string for a name'):
        validate_circuit(circuit)


def test_validate_circuit_checks_circuit_has_at_least_one_instruction(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when circuit instructions container has 0 instructions
    """
    circuit = sample_circuit.model_copy()
    circuit.instructions = tuple()
    with pytest.raises(ValueError, match='Each circuit should have at least one instruction'):
        validate_circuit(circuit)


def test_validate_circuit_checks_instruction_name_is_supported(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when instruction name is set to an unknown instruction type
    """
    circuit = sample_circuit.model_copy()
    circuit.instructions[0].name = 'kaboom'
    with pytest.raises(ValueError, match='Unknown operation "kaboom"'):
        validate_circuit(circuit)


def test_validate_circuit_checks_instruction_implementation_is_string(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when instruction implementation is set to an empty string
    """
    circuit = sample_circuit.model_copy()
    circuit.instructions[0].implementation = ''
    with pytest.raises(ValueError, match='Implementation of the instruction should be None, or a non-empty string'):
        validate_circuit(circuit)


def test_validate_circuit_checks_instruction_qubit_count(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when qubit count of the instruction does not match the arity of
    that instruction
    """
    circuit = sample_circuit.model_copy()
    circuit.instructions[0].qubits += ('Qubit C',)
    with pytest.raises(ValueError, match=r'The "cz" operation acts on 2 qubit\(s\), but 3 were given'):
        validate_circuit(circuit)


def test_validate_circuit_extra_arguments(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when submitted argument names of the instruction are not supported
    """
    circuit = sample_circuit.model_copy()
    circuit.instructions[1].args['arg_x'] = 'This argument name is not supported by the operation'
    with pytest.raises(ValueError, match='The operation "prx" allows'):
        validate_circuit(circuit)


def test_validate_circuit_missing_arguments():
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when submitted argument names of the instruction are not supported
    """
    with pytest.raises(ValueError, match='The operation "prx" requires'):
        Circuit(
            name='The circuit',
            instructions=[
                Instruction(
                    name='prx',
                    qubits=('QB1',),
                    args={'phase_t': 0.3},
                ),
            ],
        )


def test_validate_circuit_checks_instruction_argument_types(sample_circuit):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    catches when submitted argument types of the instruction are not supported
    """
    circuit = sample_circuit.model_copy()
    circuit.instructions[1].args['phase_t'] = '0.7'
    with pytest.raises(TypeError, match='The argument "phase_t" should be of one of the following supported types'):
        validate_circuit(circuit)


def test_validate_circuit_can_handle_raw_instructions(sample_circuit_with_raw_instructions):
    """
    Tests that custom Pydantic validator (triggered via <validate_circuit>)
    accepts a circuit with raw instructions
    """
    circuit = sample_circuit_with_raw_instructions.model_copy()
    validate_circuit(circuit)


def test_abort_job_successful(sample_client, existing_job_url, existing_run_id, abort_job_success):
    """
    Tests aborting a job
    """
    expect(requests, times=1).post(f'{existing_job_url}/abort', **post_jobs_args()).thenReturn(abort_job_success)

    sample_client.abort_job(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


@pytest.mark.parametrize('status_code', [404, 409])
def test_abort_job_failed(status_code, sample_client, existing_job_url, existing_run_id, abort_job_failed):
    """
    Tests aborting a job raises JobAbortionError if server returned error response
    """
    response = abort_job_failed
    response.status_code = status_code
    expect(requests, times=1).post(f'{existing_job_url}/abort', **post_jobs_args()).thenReturn(response)

    with pytest.raises(JobAbortionError):
        sample_client.abort_job(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


@pytest.mark.parametrize(
    'params',
    [
        {},
        {'options': CircuitCompilationOptions(heralding_mode=HeraldingMode.ZEROS, active_reset_cycles=1)},
        {'custom_settings': {'some_setting': 1}},
        {'calibration_set_id': uuid.uuid4()},
        {'options': CircuitCompilationOptions(max_circuit_duration_over_t2=0.0)},
        {'qubit_mapping': {'QB1': 'QB2', 'QB2': 'QB1'}},
    ],
)
def test_create_and_submit_run_request(
    sample_client,
    sample_circuit,
    jobs_url,
    submit_success,
    existing_run_id,
    dynamic_architecture_url,
    dynamic_architecture_success,
    params,
):
    """
    Tests that calling create_run_request and then submit_run_request is equivalent to calling submit_circuits.
    """
    if 'calibration_set_id' in params:
        when(requests).get(
            sample_client._api.url(APIEndpoint.CALIBRATED_GATES, params['calibration_set_id']), ...
        ).thenReturn(dynamic_architecture_success)
    else:
        when(requests).get(dynamic_architecture_url, ...).thenReturn(dynamic_architecture_success)

    run_request = sample_client.create_run_request([sample_circuit], **params)
    if 'options' in params:
        assert run_request.active_reset_cycles == params['options'].active_reset_cycles
    expect(requests, times=2).post(jobs_url, **post_jobs_args(run_request)).thenReturn(submit_success)
    assert sample_client.submit_run_request(run_request) == existing_run_id
    assert sample_client.submit_circuits([sample_circuit], **params) == existing_run_id

    verifyNoUnwantedInteractions()
    unstub()


@pytest.mark.parametrize(
    'run_request_name, quantum_architecture_name, sample_circuit_name',
    [
        (run_request, success_result, sample_circuit)
        for run_request in [
            'run_request_with_move_validation',
            'run_request_without_prx_move_validation',
            'run_request_with_move_gate_frame_tracking',
        ]
        for success_result, sample_circuit in zip(
            ['dynamic_architecture_success', 'move_architecture_success', 'move_architecture_success'],
            ['sample_circuit', 'move_circuit', 'move_circuit_with_prx_in_the_sandwich'],
        )
    ],
)
def test_compiler_options_are_used_and_sent(
    sample_client,
    sample_circuit_name,
    jobs_url,
    run_request_name,
    request,
    submit_success,
    dynamic_architecture_url,
    quantum_architecture_name,
):
    """
    Tests submitting circuits for execution
    """
    run_request = request.getfixturevalue(run_request_name)
    quantum_architecture_success = request.getfixturevalue(quantum_architecture_name)
    run_request.circuits = [request.getfixturevalue(sample_circuit_name)]

    when(requests).get(dynamic_architecture_url, ...).thenReturn(quantum_architecture_success)
    if (
        sample_circuit_name != 'move_circuit_with_prx_in_the_sandwich'  # Valid circuit
        or run_request_name == 'run_request_without_prx_move_validation'  # Validation is turned off
    ):
        expect(requests, times=1).post(jobs_url, **post_jobs_args(run_request)).thenReturn(submit_success)
        sample_client.submit_circuits(**submit_circuits_args(run_request))
    else:  # Invalid circuit and validation is turned on.
        with pytest.raises(CircuitValidationError):
            sample_client.submit_circuits(**submit_circuits_args(run_request))

    verifyNoUnwantedInteractions()
    unstub()


def test_get_dynamic_quantum_architecture_with_calset_id(
    sample_client, base_url, dynamic_architecture_success, sample_dynamic_architecture
):
    """Tests that the correct dynamic quantum architecture for the given ``calibration_set_id`` is returned."""
    calset_id = sample_dynamic_architecture.calibration_set_id
    expect(requests, times=1).get(f'{base_url}/api/v1/calibration/{calset_id}/gates', ...).thenReturn(
        dynamic_architecture_success
    )
    assert sample_client.get_dynamic_quantum_architecture(calset_id) == sample_dynamic_architecture
    verifyNoUnwantedInteractions()
    unstub()


def test_get_dynamic_quantum_architecture_with_calset_id_caches(
    sample_client, base_url, dynamic_architecture_success, sample_dynamic_architecture
):
    """
    Tests that cached dynamic quantum architecture is returned when requesting it for the second time for
    a given calibration set id.
    """
    calset_id = sample_dynamic_architecture.calibration_set_id
    expect(requests, times=1).get(f'{base_url}/api/v1/calibration/{calset_id}/gates', ...).thenReturn(
        dynamic_architecture_success
    )

    assert sample_client.get_dynamic_quantum_architecture(calset_id) == sample_dynamic_architecture
    assert sample_client.get_dynamic_quantum_architecture(calset_id) == sample_dynamic_architecture

    verifyNoUnwantedInteractions()
    unstub()


def test_get_dynamic_quantum_architecture_without_calset_id(
    sample_client, base_url, dynamic_architecture_success, sample_dynamic_architecture
):
    """Tests that the correct dynamic quantum architecture for the default calibration set is returned."""
    expect(requests, times=1).get(f'{base_url}/api/v1/calibration/default/gates', ...).thenReturn(
        dynamic_architecture_success
    )
    assert sample_client.get_dynamic_quantum_architecture() == sample_dynamic_architecture
    verifyNoUnwantedInteractions()
    unstub()


def test_get_dynamic_quantum_architecture_without_calset_id_does_not_cache(
    sample_client, base_url, dynamic_architecture_success, sample_dynamic_architecture
):
    """
    Tests that the correct dynamic quantum architecture is returned in the case where default calset
    changes between two invocations of get_dynamic_quantum_architecture().
    """
    dynamic_quantum_architecture_2 = {
        'calibration_set_id': '3902d525-d8f4-42c0-9fa9-6bbd535b6c80',
        'qubits': ['QB1', 'QB2'],
        'computational_resonators': [],
        'gates': {
            'prx': {
                'implementations': {
                    'drag_gaussian': {
                        'loci': [['QB1'], ['QB2']],
                    }
                },
                'default_implementation': 'drag_gaussian',
                'override_default_implementation': {},
            },
        },
    }
    dynamic_architecture_success_2 = MockJsonResponse(200, dynamic_quantum_architecture_2)
    expect(requests, times=2).get(f'{base_url}/api/v1/calibration/default/gates', ...).thenReturn(
        dynamic_architecture_success
    ).thenReturn(dynamic_architecture_success_2)

    assert sample_client.get_dynamic_quantum_architecture() == sample_dynamic_architecture
    assert sample_client.get_dynamic_quantum_architecture() == DynamicQuantumArchitecture(
        **dynamic_quantum_architecture_2
    )

    verifyNoUnwantedInteractions()
    unstub()


def test_get_dynamic_quantum_architecture_throws_json_decode_error_if_received_not_json(
    sample_client, base_url, not_valid_json_response
):
    """Test that an exception is raised when the response is not a valid JSON"""
    expect(requests, times=1).get(f'{base_url}/api/v1/calibration/default/gates', ...).thenReturn(
        not_valid_json_response
    )

    with pytest.raises(ArchitectureRetrievalError):
        sample_client.get_dynamic_quantum_architecture()

    verifyNoUnwantedInteractions()
    unstub()


def test_get_dynamic_quantum_architecture_not_found(base_url, sample_client):
    """Test that an informative error message is returned when 404 is returned due to version incompatibility."""
    client_version = parse(version('iqm-client'))
    min_version = f'{client_version.major + 2}.0'
    max_version = f'{client_version.major + 3}.0'
    when(requests).get(f'{base_url}/info/client-libraries', headers=ANY, timeout=ANY).thenReturn(
        mock_supported_client_libraries_response(min_version=min_version, max_version=max_version)
    )
    when(requests).get(f'{base_url}/api/v1/calibration/default/gates', ...).thenReturn(MockJsonResponse(404, {}))
    with pytest.raises(
        HTTPError,
        match=re.escape(
            f'Your IQM Client version {client_version} was built for a different version of IQM Server. '
            f'You might encounter issues. For the best experience, consider using a version '
            f'of IQM Client that satisfies {min_version} <= iqm-client < {max_version}.'
        ),
    ):
        sample_client.get_dynamic_quantum_architecture()
    unstub()


@pytest.mark.parametrize(
    'iqm_client_name,server_version_diff',
    [
        ('iqm_client', 0),
        ('iqm-client', 1),
    ],
)
def test_check_versions_success(base_url, iqm_client_name, server_version_diff, recwarn):
    """Test that a warning about version incompatibility is shown when initializing client with incompatible server."""
    client_version = parse(version('iqm-client'))
    min_version = f'{client_version.major + server_version_diff}.0'
    max_version = f'{client_version.major + server_version_diff + 1}.0'
    when(requests).get(f'{base_url}/info/client-libraries', headers=ANY, timeout=ANY).thenReturn(
        mock_supported_client_libraries_response(
            iqm_client_name=iqm_client_name, min_version=min_version, max_version=max_version
        )
    )
    if server_version_diff == 0:
        IQMClient(base_url)
        assert len(recwarn) == 0
    else:
        with pytest.warns(
            UserWarning,
            match=re.escape(
                f'Your IQM Client version {client_version} was built for a different version of IQM Server. '
                f'You might encounter issues. For the best experience, consider using a version '
                f'of IQM Client that satisfies {min_version} <= iqm-client < {max_version}.'
            ),
        ):
            IQMClient(base_url)
    unstub()


def test_check_versions_bad_response(base_url):
    """Test that unexpected response from /client-libraries endpoint does not break the client."""
    when(requests).get(f'{base_url}/info/client-libraries', headers=ANY, timeout=ANY).thenReturn(
        MockJsonResponse(
            200,
            {'iqm_client': 'invalid-payload'},
        )
    )
    with pytest.warns(
        UserWarning,
        match=re.escape('Could not verify IQM Client compatibility with the server. You might encounter issues.'),
    ):
        IQMClient(base_url)
    unstub()


def test_check_versions_request_exception(base_url):
    """Test that exception raised by request to /client-libraries endpoint does not break the client."""
    when(requests).get(f'{base_url}/info/client-libraries', headers=ANY, timeout=ANY).thenRaise(
        requests.exceptions.RequestException
    )
    with pytest.warns(
        UserWarning,
        match=re.escape('Could not verify IQM Client compatibility with the server. You might encounter issues.'),
    ):
        IQMClient(base_url)
    unstub()


def test_get_run_counts(base_url, sample_client, existing_run_id):
    """Test that the number of runs is returned."""
    expect(requests, times=1).get(f'{base_url}/jobs/{existing_run_id}/counts', ...).thenReturn(
        MockJsonResponse(
            200,
            {
                'status': 'ready',
                'counts_batch': [
                    {
                        'measurement_keys': ['m1'],
                        'counts': {'0': 5, '1': 5},
                    },
                    {
                        'measurement_keys': ['m2'],
                        'counts': {'0': 1, '1': 9},
                    },
                ],
                'warnings': [],
            },
        )
    )
    counts = sample_client.get_run_counts(existing_run_id)
    assert counts.status == Status.READY
    assert counts.counts_batch == [
        Counts(measurement_keys=['m1'], counts={'0': 5, '1': 5}),
        Counts(measurement_keys=['m2'], counts={'0': 1, '1': 9}),
    ]
    verifyNoUnwantedInteractions()
    unstub()


def test_get_supported_client_libraries(base_url, sample_client):
    """Test retrieving client library information from server."""
    libraries_data = {
        'iqm-client': {
            'name': 'IQM Client',
            'package_name': 'iqm-client',
            'repo_url': 'https://github.com/iqm-finland/iqm-client',
            'package_url': 'https://pypi.org/project/iqm-client',
            'min': '14.0.0',
            'max': '15.0.0',
            'images': None,
        },
        'iqm-cortex-cli': {
            'name': 'IQM Cortex CLI',
            'package_name': 'iqm-cortex-cli',
            'repo_url': 'https://github.com/iqm-finland/cortex-cli',
            'package_url': 'https://pypi.org/project/iqm-cortex-cli',
            'min': '1.0.0',
            'max': '2.0.0',
            'images': [],
        },
    }

    expect(requests, times=1).get(f'{base_url}/info/client-libraries', headers=ANY, timeout=ANY).thenReturn(
        MockJsonResponse(200, libraries_data)
    )

    libraries = sample_client.get_supported_client_libraries()

    # Verify returned data matches expected structure
    assert len(libraries) == 2
    assert 'iqm-client' in libraries
    assert 'iqm-cortex-cli' in libraries

    # Verify specific library details
    iqm_client = libraries['iqm-client']
    assert iqm_client.name == 'IQM Client'
    assert iqm_client.package_name == 'iqm-client'
    assert iqm_client.repo_url == 'https://github.com/iqm-finland/iqm-client'
    assert iqm_client.package_url == 'https://pypi.org/project/iqm-client'
    assert iqm_client.min == '14.0.0'
    assert iqm_client.max == '15.0.0'

    verifyNoUnwantedInteractions()
    unstub()



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=too-many-lines
"""This module contains the data models used by IQMClient."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
import re
from typing import Any, Final, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, StrictStr, field_validator
from pydantic_core.core_schema import ValidationInfo


@dataclass(frozen=True)
class NativeOperation:
    """Describes a native operation on the quantum computer."""

    # pylint: disable=too-many-instance-attributes

    name: str
    """Name of the operation."""
    arity: int
    """Number of locus components (usually qubits) the operation acts on.
    Zero means the operation can be applied on any number of locus components."""
    args_required: dict[str, tuple[type, ...]] = field(default_factory=dict)
    """Maps names of required operation parameters to their allowed types."""
    args_not_required: dict[str, tuple[type, ...]] = field(default_factory=dict)
    """Maps names of optional operation parameters to their allowed types."""
    symmetric: bool = False
    """True iff the effect of operation is symmetric in the locus components it acts on.
    Only meaningful if :attr:`arity` != 1."""
    renamed_to: str = ''
    """If nonempty, indicates that this operation name is deprecated, and IQM client will
    auto-rename it to the new name."""
    factorizable: bool = False
    """Iff True, any multi-component instance of this operation can be broken down to
    single-component instances, and calibration data is specific to single-component loci."""
    no_calibration_needed: bool = False
    """Iff true, the operation is always allowed on all QPU loci regardless of calibration state.
    Typically a metaoperation like barrier."""


_SUPPORTED_OPERATIONS: dict[str, NativeOperation] = {
    op.name: op
    for op in [
        NativeOperation('barrier', 0, symmetric=True, no_calibration_needed=True),
        NativeOperation('measure', 0, {'key': (str,)}, {'feedback_key': (str,)}, factorizable=True),
        NativeOperation(
            'prx',
            1,
            {
                'angle_t': (float, int),
                'phase_t': (float, int),
            },
        ),
        NativeOperation(
            'cc_prx',
            1,
            {
                'angle_t': (float, int),
                'phase_t': (float, int),
                'feedback_key': (str,),
                'feedback_qubit': (str,),
            },
        ),
        NativeOperation('cz', 2, symmetric=True),
        NativeOperation('move', 2),
    ]
}

Locus = tuple[StrictStr, ...]
"""Names of the QPU components (typically qubits) a quantum operation instance is acting on, e.g. `("QB1", "QB2")`."""


class Instruction(BaseModel):
    r"""Native quantum operation instance with particular arguments and locus.

    This class represents a native quantum operation
    acting on :attr:`qubits`, with the arguments :attr:`args`.
    The operation is determined by :attr:`name`.

    We currently support the following native operations:

    ================ =========== ======================================= ===========
    name             # of qubits args                                    description
    ================ =========== ======================================= ===========
    measure          >= 1        ``key: str``, ``feedback_key: str``     Measurement in the Z basis.
    prx              1           ``angle_t: float``, ``phase_t: float``  Phased x-rotation gate.
    cc_prx           1           ``angle_t: float``, ``phase_t: float``,
                                 ``feedback_qubit: str``,
                                 ``feedback_key: str``                   Classically controlled PRX gate.
    cz               2                                                   Controlled-Z gate.
    move             2                                                   Moves a qubit state between a qubit and a
                                                                         computational resonator, as long as
                                                                         at least one of the components is
                                                                         in the :math:`|0\rangle` state.
    barrier          >= 1                                                Execution barrier.
    ================ =========== ======================================= ===========

    For each Instruction you may also optionally specify :attr:`~Instruction.implementation`,
    which contains the name of an implementation of the operation to use.
    Support for multiple implementations is currently experimental and in normal use the
    field should be omitted, this selects the default implementation for the operation for that locus.

    Measure
    -------

    Measurement in the computational (Z) basis. The measurement results are the output of the circuit.
    Takes two string arguments: ``key``, denoting the measurement key the returned results are labeled with,
    and ``feedback_key``, which is only needed if the measurement result is used for classical control
    within the circuit.
    All the measurement keys and feedback keys used in a circuit must be unique (but the two groups of
    keys are independent namespaces).
    Each qubit may be measured multiple times, i.e. mid-circuit measurements are allowed.

    .. code-block:: python
        :caption: Example

        Instruction(name='measure', qubits=('alice', 'bob', 'charlie'), args={'key': 'm1'})

    PRX
    ---

    Phased x-rotation gate, i.e. an x-rotation conjugated by a z-rotation.
    Takes two arguments, the rotation angle ``angle_t`` and the phase angle ``phase_t``,
    both measured in units of full turns (:math:`2\pi` radians).
    The gate is represented in the standard computational basis by the matrix

    .. math::
        \text{PRX}(\theta, \phi) = \exp(-i (X \cos (2 \pi \; \phi) + Y \sin (2 \pi \; \phi)) \: \pi \; \theta)
        = \text{RZ}(\phi) \: \text{RX}(\theta) \: \text{RZ}^\dagger(\phi),

    where :math:`\theta` = ``angle_t``, :math:`\phi` = ``phase_t``,
    and :math:`X` and :math:`Y` are Pauli matrices.

    .. code-block:: python
        :caption: Example

        Instruction(name='prx', qubits=('bob',), args={'angle_t': 0.7, 'phase_t': 0.25})

    CC_PRX
    ------

    Classically controlled PRX gate. Takes four arguments. ``angle_t`` and ``phase_t`` are exactly as in PRX.
    ``feedback_key`` is a string that identifies the ``measure`` instruction whose result controls
    the gate (the one that shares the feedback key).
    ``feedback_qubit`` is the name of the physical qubit within the ``measure`` instruction that produces the feedback.
    If the measurement result is 1, the PRX gate is applied. If it is 0, an identity gate of similar time
    duration gate is applied instead.
    The measurement instruction must precede the classically controlled gate instruction in the quantum circuit.

    CZ
    --

    Controlled-Z gate. Represented in the standard computational basis by the matrix

    .. math:: \text{CZ} = \text{diag}(1, 1, 1, -1).

    It is symmetric wrt. the qubits it's acting on, and takes no arguments.

    .. code-block:: python
        :caption: Example

        Instruction(name='cz', qubits=('alice', 'bob'), args={})

    MOVE
    ----

    The MOVE operation is a unitary population exchange operation between a qubit and a resonator.
    Its effect is only defined in the invariant subspace :math:`S = \text{span}\{|00\rangle, |01\rangle, |10\rangle\}`,
    where it swaps the populations of the states :math:`|01\rangle` and :math:`|10\rangle`.
    Its effect on the orthogonal subspace is undefined.

    MOVE has the following presentation in the subspace :math:`S`:

    .. math:: \text{MOVE}_S = |00\rangle \langle 00| + a |10\rangle \langle 01| + a^{-1} |01\rangle \langle 10|,

    where :math:`a` is an undefined complex phase that is canceled when the MOVE gate is applied a second time.

    To ensure that the state of the qubit and resonator has no overlap with :math:`|11\rangle`, it is
    recommended that no single qubit gates are applied to the qubit in between a
    pair of MOVE operations.

    .. code-block:: python
        :caption: Example

        Instruction(name='move', qubits=('alice', 'resonator'), args={})

    .. note:: MOVE is only available in quantum computers with the IQM Star architecture.

    Barrier
    -------

    A barrier instruction affects the physical execution order of the instructions elsewhere in the
    circuit that act on qubits spanned by the barrier.
    It ensures that any such instructions that succeed the barrier are only executed after
    all such instructions that precede the barrier have been completed.
    Hence it can be used to guarantee a specific causal order for the other instructions.
    It takes no arguments, and has no other effect.

    .. code-block:: python
        :caption: Example

        Instruction(name='barrier', qubits=('alice', 'bob'), args={})

    .. note::

       One-qubit barriers will not have any effect on circuit's compilation and execution. Higher layers
       that sit on top of IQM Client can make actual use of one-qubit barriers (e.g. during circuit optimization),
       therefore having them is allowed.
    """

    name: str = Field(..., examples=['measure'])
    """name of the quantum operation"""
    implementation: Optional[StrictStr] = Field(None)
    """name of the implementation, for experimental use only"""
    qubits: Locus = Field(..., examples=[('alice',)])
    """names of the locus components (typically qubits) the operation acts on"""
    args: dict[str, Any] = Field(..., examples=[{'key': 'm'}])
    """arguments for the operation"""

    def __init__(self, **data):
        super().__init__(**data)
        # Auto-convert name if a deprecated name is used
        self.name = _op_current_name(self.name)

    @field_validator('name')
    @classmethod
    def name_validator(cls, value):
        """Check if the name of instruction is set to one of the supported quantum operations."""
        name = value
        if name not in _SUPPORTED_OPERATIONS:
            message = ', '.join(_SUPPORTED_OPERATIONS)
            raise ValueError(f'Unknown operation "{name}". Supported operations are "{message}"')
        return name

    @field_validator('implementation')
    @classmethod
    def implementation_validator(cls, value):
        """Check if the implementation of the instruction is set to a non-empty string."""
        implementation = value
        if isinstance(implementation, str):
            if not implementation:
                raise ValueError('Implementation of the instruction should be None, or a non-empty string')
        return implementation

    @field_validator('qubits')
    @classmethod
    def qubits_validator(cls, value, info: ValidationInfo):
        """Check if the instruction has the correct number of qubits for its operation."""
        qubits = value
        name = info.data.get('name')
        if not name:
            raise ValueError('Could not validate qubits because the name of the instruction did not pass validation')
        arity = _SUPPORTED_OPERATIONS[name].arity
        if (0 < arity) and (arity != len(qubits)):
            raise ValueError(f'The "{name}" operation acts on {arity} qubit(s), but {len(qubits)} were given: {qubits}')
        return qubits

    @field_validator('args')
    @classmethod
    def args_validator(cls, value, info: ValidationInfo):
        """Check argument names and types for a given instruction"""
        args = value
        name = info.data.get('name')
        if not name:
            raise ValueError('Could not validate args because the name of the instruction did not pass validation')

        # Check argument names
        submitted_arg_names = set(args)
        required_arg_names = set(_SUPPORTED_OPERATIONS[name].args_required)
        allowed_arg_types = _SUPPORTED_OPERATIONS[name].args_required | _SUPPORTED_OPERATIONS[name].args_not_required
        allowed_arg_names = set(allowed_arg_types)
        if not required_arg_names <= submitted_arg_names:
            raise ValueError(
                f'The operation "{name}" requires '
                f'{tuple(required_arg_names)} argument(s), '
                f'but {tuple(submitted_arg_names)} were given'
            )
        if not submitted_arg_names <= allowed_arg_names:
            message = tuple(allowed_arg_names) if allowed_arg_names else 'no'
            raise ValueError(
                f'The operation "{name}" allows '
                f'{message} argument(s), '
                f'but {tuple(submitted_arg_names)} were given'
            )

        # Check argument types
        for arg_name, arg_value in args.items():
            allowed_types = allowed_arg_types[arg_name]
            if not isinstance(arg_value, allowed_types):
                raise TypeError(
                    f'The argument "{arg_name}" should be of one of the following supported types'
                    f' {allowed_types}, but ({type(arg_value)}) was given'
                )

        return value


def _op_is_symmetric(name: str) -> bool:
    """Returns True iff the given native operation is symmetric, i.e. the order of the
    locus components does not matter.

    Args:
        name: name of the operation
    Returns:
        True iff the locus order does not matter
    Raises:
        KeyError: ``name`` is unknown
    """
    return _SUPPORTED_OPERATIONS[name].symmetric


def _op_arity(name: str) -> int:
    """Returns the arity of the given native operation, i.e. the number of locus components it acts on.

    Zero means any number of locus components is OK.

    Args:
        name: name of the operation
    Returns:
        arity of the operation
    Raises:
        KeyError: ``name`` is unknown
    """
    return _SUPPORTED_OPERATIONS[name].arity


def _op_current_name(name: str) -> str:
    """Checks if the operation name has been deprecated and returns the new name if it is;
    otherwise, just returns the name as-is.

    Args:
        name: name of the operation

    Returns:
        current name of the operation
    Raises:
        KeyError: ``name`` is unknown
    """
    return _SUPPORTED_OPERATIONS[name].renamed_to or name


class Circuit(BaseModel):
    """Quantum circuit to be executed.

    Consists of native quantum operations, each represented by an instance of the :class:`Instruction` class.
    """

    name: str = Field(..., examples=['test circuit'])
    """name of the circuit"""
    instructions: tuple[Instruction, ...] = Field(...)
    """instructions comprising the circuit"""
    metadata: Optional[dict[str, Any]] = Field(None)
    """arbitrary metadata associated with the circuit"""

    def all_qubits(self) -> set[str]:
        """Return the names of all qubits in the circuit."""
        qubits: set[str] = set()
        for instruction in self.instructions:
            qubits.update(instruction.qubits)
        return qubits

    @field_validator('name')
    @classmethod
    def name_validator(cls, value):
        """Check if the circuit name is a non-empty string"""
        name = value
        if len(name) == 0:
            raise ValueError('A circuit should have a non-empty string for a name.')
        return name

    @field_validator('instructions')
    @classmethod
    def instructions_validator(cls, value):
        """Check the container of instructions and each instruction within"""
        instructions = value

        # Check container type
        if not isinstance(instructions, (list, tuple)):
            raise ValueError('Instructions of a circuit should be packed in a tuple')

        # Check if any instructions are present
        if len(value) == 0:
            raise ValueError('Each circuit should have at least one instruction.')

        # Check each instruction explicitly, because automatic validation for Instruction
        # is only called when we create a new instance of Instruction, but not if we modify
        # an existing instance.
        for instruction in instructions:
            if isinstance(instruction, Instruction):
                Instruction.model_validate(instruction.__dict__)
            else:
                raise ValueError('Every instruction in a circuit should be of type <Instruction>')

        return instructions


CircuitBatch = list[Circuit]
"""Type that represents a list of quantum circuits to be executed together in a single batch."""


def validate_circuit(circuit: Circuit) -> None:
    """Validates a submitted quantum circuit using Pydantic tooling.

    Args:
        circuit: a circuit that needs validation

    Raises:
        pydantic.error_wrappers.ValidationError: validation failed
    """
    Circuit.model_validate(circuit.__dict__)


class SingleQubitMapping(BaseModel):
    """Mapping of a logical qubit name to a physical qubit name."""

    logical_name: str = Field(..., examples=['alice'])
    """logical qubit name"""
    physical_name: str = Field(..., examples=['QB1'])
    """physical qubit name"""


QubitMapping = list[SingleQubitMapping]
"""Type that represents a qubit mapping for a circuit, i.e. a list of single qubit mappings
for all qubits in the circuit."""


def serialize_qubit_mapping(qubit_mapping: dict[str, str]) -> list[SingleQubitMapping]:
    """Serializes a qubit mapping dict into the corresponding IQM data transfer format.

    Args:
        qubit_mapping: mapping from logical to physical qubit names

    Returns:
        data transfer object representing the mapping
    """
    return [SingleQubitMapping(logical_name=k, physical_name=v) for k, v in qubit_mapping.items()]


class QuantumArchitectureSpecification(BaseModel):
    """Quantum architecture specification."""

    name: str = Field(...)
    """Name of the quantum architecture."""
    operations: dict[str, list[list[str]]] = Field(...)
    """Operations supported by this quantum architecture, mapped to the allowed loci."""
    qubits: list[str] = Field(...)
    """List of qubits of this quantum architecture."""
    qubit_connectivity: list[list[str]] = Field(...)
    """Qubit connectivity of this quantum architecture."""

    def __init__(self, **data):
        operations = data.get('operations')
        if isinstance(operations, list):
            # backwards compatibility for the old quantum architecture format
            qubits = data.get('qubits')
            qubit_connectivity = data.get('qubit_connectivity')
            # add all possible loci for the ops
            data['operations'] = {
                _op_current_name(op): (
                    qubit_connectivity if _op_arity(_op_current_name(op)) == 2 else [[qb] for qb in qubits]
                )
                for op in operations
            }

        super().__init__(**data)
        self.operations = {_op_current_name(k): v for k, v in self.operations.items()}

    def has_equivalent_operations(self, other: QuantumArchitectureSpecification) -> bool:
        """Compares the given operation sets defined by the quantum architecture against
        another architecture specification.

        Returns:
            True if the operation and the loci are equivalent.
        """
        return QuantumArchitectureSpecification.compare_operations(self.operations, other.operations)

    @staticmethod
    def compare_operations(ops1: dict[str, list[list[str]]], ops2: dict[str, list[list[str]]]) -> bool:
        """Compares the given operation sets.

        Returns:
            True if the operation and the loci are equivalent.
        """
        if set(ops1) != set(ops2):
            return False
        for op, loci1 in ops1.items():
            loci2 = ops2[op]
            if _op_is_symmetric(op):
                # for comparing symmetric instruction loci, sorting order does not matter as long as it's consistent
                l1 = [tuple(sorted(locus)) for locus in loci1]
                l2 = [tuple(sorted(locus)) for locus in loci2]
            else:
                l1 = [tuple(locus) for locus in loci1]
                l2 = [tuple(locus) for locus in loci2]

            if set(l1) != set(l2):
                return False
        return True


class QuantumArchitecture(BaseModel):
    """Quantum architecture as returned by server."""

    quantum_architecture: QuantumArchitectureSpecification = Field(...)
    """Details about the quantum architecture."""


def _component_sort_key(component_name: str) -> tuple[str, int, str]:
    def get_numeric_id(name: str) -> int:
        match = re.search(r'(\d+)', name)
        return int(match.group(1)) if match else 0

    return re.sub(r'[^a-zA-Z]', '', component_name), get_numeric_id(component_name), component_name


class GateImplementationInfo(BaseModel):
    """Information about an implementation of a quantum gate/operation."""

    loci: tuple[Locus, ...] = Field(...)
    """loci for which this gate implementation has been calibrated"""


class GateInfo(BaseModel):
    """Information about a quantum gate/operation."""

    implementations: dict[str, GateImplementationInfo] = Field(...)
    """mapping of available implementation names to information about the implementations"""
    default_implementation: str = Field(...)
    """default implementation for the gate, used unless overridden by :attr:`override_default_implementation`
    or unless the user requests a specific implementation for a particular gate in the circuit using
    :attr:`.Instruction.implementation`"""
    override_default_implementation: dict[Locus, str] = Field(...)
    """mapping of loci to implementation names that override ``default_implementation`` for those loci"""

    @cached_property
    def loci(self) -> tuple[Locus, ...]:
        """Returns all loci which are available for at least one of the implementations.

        The loci are sorted first based on the first locus component, then the second, etc.
        The sorting of individual locus components is first done alphabetically based on their
        non-numeric part, and then components with the same non-numeric part are sorted numerically.
        An example of loci sorted this way would be:

            ('QB1', 'QB2'),
            ('QB2', 'COMPR1'),
            ('QB2', 'QB3'),
            ('QB3', 'COMPR1'),
            ('QB3', 'COMPR2'),
            ('QB3', 'QB1'),
            ('QB10', 'QB2'),

        """
        loci_set = set(locus for impl in self.implementations.values() for locus in impl.loci)
        loci_sorted = sorted(loci_set, key=lambda locus: tuple(map(_component_sort_key, locus)))
        return tuple(loci_sorted)


class DynamicQuantumArchitecture(BaseModel):
    """Dynamic quantum architecture as returned by server.

    The dynamic quantum architecture (DQA) describes gates/operations for which calibration data
    exists in the calibration set.
    """

    calibration_set_id: UUID = Field(...)
    """id of the calibration set from which this DQA was generated"""
    qubits: list[str] = Field(...)
    """qubits that appear in at least one gate locus in the calibration set"""
    computational_resonators: list[str] = Field(...)
    """computational resonators that appear in at least one gate locus in the calibration set"""
    gates: dict[str, GateInfo] = Field(...)
    """mapping of gate names to information about the gates"""

    @cached_property
    def components(self) -> tuple[str, ...]:
        """Returns all locus components (qubits and computational resonators) sorted.

        The components are first sorted alphabetically based on their non-numeric part, and then
        components with the same non-numeric part are sorted numerically. An example of components
        sorted this way would be: ('COMPR1', 'COMPR2', 'QB1', 'QB2', 'QB3', 'QB10', 'QB11', 'QB20').
        """
        return tuple(sorted(self.qubits + self.computational_resonators, key=_component_sort_key))


class HeraldingMode(str, Enum):
    """Heralding mode for circuit execution.

    Heralding is the practice of generating data about the state of qubits prior to execution of a circuit.
    This can be achieved by measuring the qubits immediately before executing each shot for a circuit."""

    NONE = 'none'
    """Do not do any heralding."""
    ZEROS = 'zeros'
    """Perform a heralding measurement after qubit initialization, only retain shots with an all-zeros result.

    Note: in this mode, the number of shots returned after execution will be less or equal to the requested amount
    due to the post-selection based on heralding data."""


class MoveGateValidationMode(str, Enum):
    """MOVE gate validation mode for circuit compilation. This options is meant for advanced users."""

    STRICT: Final[str] = 'strict'
    """Perform standard MOVE gate validation: MOVE gates must only appear in sandwiches, with no gates acting on the
    MOVE qubit inside the sandwich."""
    ALLOW_PRX: Final[str] = 'allow_prx'
    """Allow PRX gates on the MOVE qubit inside MOVE sandwiches during validation."""
    NONE: Final[str] = 'none'
    """Do not perform any MOVE gate validation."""


class MoveGateFrameTrackingMode(str, Enum):
    """MOVE gate frame tracking mode for circuit compilation. This option is meant for advanced users."""

    FULL: Final[str] = 'full'
    """Perform complete MOVE gate frame tracking."""
    NO_DETUNING_CORRECTION: Final[str] = 'no_detuning_correction'
    """Do not add the phase detuning corrections to the pulse schedule for the MOVE gate. The user is expected to do
    these manually."""
    NONE: Final[str] = 'none'
    """Do not perform any MOVE gate frame tracking. The user is expected to do these manually."""


class DDMode(str, Enum):
    """Dynamical Decoupling (DD) mode for circuit execution."""

    DISABLED: Final[str] = 'disabled'
    """Do not apply dynamical decoupling."""
    ENABLED: Final[str] = 'enabled'
    """Apply dynamical decoupling."""


PRXSequence = list[tuple[float, float]]
"""A sequence of PRX gates. A generic PRX gate is defined by rotation angle and phase angle, Theta and Phi,
respectively."""


@dataclass
class DDStrategy:
    """Describes a particular dynamical decoupling strategy.

    The current standard DD stategy can be found in :attr:`.STANDARD_DD_STRATEGY`,
    but users can use this class to provide their own dynamical decoupling strategies.

    See Ezzell et al., Phys. Rev. Appl. 20, 064027 (2022) for information on DD sequences.
    """

    merge_contiguous_waits: bool = True
    """Merge contiguous ``Wait`` instructions into one if they are separated only by ``Block`` instructions."""

    target_qubits: Optional[frozenset[str]] = None
    """Qubits on which dynamical decoupling should be applied. If ``None``, all qubits are targeted."""

    skip_leading_wait: bool = True
    """Skip processing leading ``Wait`` instructions."""

    skip_trailing_wait: bool = True
    """Skip processing trailing ``Wait`` instructions."""

    gate_sequences: list[tuple[int, Union[str, PRXSequence], str]] = field(default_factory=list)
    """Available decoupling gate sequences to chose from in this strategy.

    Each sequence is defined by a tuple of ``(ratio, gate pattern, align)``:

        * ratio: Minimal duration for the sequence (in PRX gate durations) as integer.

        * gate pattern: Gate pattern can be defined in two ways. It can be a string containing "X" and "Y" characters,
          encoding a PRX gate sequence. For example, "YXYX" corresponds to the
          XY4 sequence, "XYXYYXYX" to the EDD sequence, etc. If more flexibility is needed, a gate pattern can be
          defined as a sequence of PRX gate argument tuples (that contain a rotation angle and a phase angle). For
          example, sequence "YX" could be written as ``[(math.pi, math.pi / 2), (math.pi, 0)]``.

        * align: Controls the alignment of the sequence within the time window it is inserted in. Supported values:

          - "asap": Corresponds to a ASAP-aligned sequence with no waiting time before the first pulse.
          - "center": Corresponds to a symmetric sequence.
          - "alap": Corresponds to a ALAP-aligned sequence.

    The Dynamical Decoupling algorithm uses the best fitting gate sequence by first sorting them
    by ``ratio`` in descending order. Then the longest fitting pattern is determined by comparing ``ratio``
    with the duration of the time window divided by the PRX gate duration.
    """


STANDARD_DD_STRATEGY = DDStrategy(gate_sequences=[(9, 'XYXYYXYX', 'asap'), (5, 'YXYX', 'asap'), (2, 'XX', 'center')])
"""The default DD strategy uses the following gate sequences:

* Simple symmetric CPMG sequence for short idling times.
* Asymmetric (left-aligned) universal XY4 sequence for medium idling times.
* Asymmetric (left-aligned) universal EDD sequence for longer idling times.
"""


@dataclass(frozen=True)
class CircuitCompilationOptions:
    """Various discrete options for quantum circuit compilation to pulse schedule."""

    max_circuit_duration_over_t2: Optional[float] = None
    """Server-side circuit disqualification threshold.
    The job is rejected on the server if any circuit in it is estimated to take longer than
    the shortest T2 time of any qubit used in the circuit, multiplied by this value.
    Setting this value to ``0.0`` turns off circuit duration checking.
    ``None`` tells the server to use its default value in the check."""
    heralding_mode: HeraldingMode = HeraldingMode.NONE
    """Heralding mode to use during the execution."""
    move_gate_validation: MoveGateValidationMode = MoveGateValidationMode.STRICT
    """MOVE gate validation mode for circuit compilation. This options is ignored on devices that do not support MOVE
        and for circuits that do not contain MOVE gates."""
    move_gate_frame_tracking: MoveGateFrameTrackingMode = MoveGateFrameTrackingMode.FULL
    """MOVE gate frame tracking mode for circuit compilation. This options is ignored on devices that do not support
        MOVE and for circuits that do not contain MOVE gates."""
    active_reset_cycles: Optional[int] = None
    """Number of active ``reset`` operations inserted at the beginning of each circuit for each active qubit.
    ``None`` means active reset is not used but instead reset is done by waiting (relaxation). Integer values smaller
    than 1 result in neither active nor reset by wait being used, in which case any reset operations must be explicitly
    added in the circuit."""
    dd_mode: DDMode = DDMode.DISABLED
    """Control whether dynamical decoupling should be enabled or disabled during the execution."""
    dd_strategy: Optional[DDStrategy] = None
    """A particular dynamical decoupling strategy to be used during the execution."""

    def __post_init__(self):
        """Validate the options."""
        if self.move_gate_frame_tracking == MoveGateFrameTrackingMode.FULL and self.move_gate_validation not in [
            MoveGateValidationMode.STRICT,
            MoveGateValidationMode.ALLOW_PRX,
            None,
        ]:
            raise ValueError(
                'Unable to perform full MOVE gate frame tracking if MOVE gate validation is not'
                + ' "strict" or "allow_prx".'
            )


class RunRequest(BaseModel):
    """Request for an IQM quantum computer to run a job that executes a batch of quantum circuits.

    Note: all circuits in a batch must measure the same qubits otherwise batch execution fails.
    """

    circuits: CircuitBatch = Field(...)
    """batch of quantum circuit(s) to execute"""
    custom_settings: Optional[dict[str, Any]] = Field(None)
    """Custom settings to override default IQM hardware settings and calibration data.
    Note: This field should be always None in normal use."""
    calibration_set_id: Optional[UUID] = Field(None)
    """ID of the calibration set to use, or None to use the latest calibration set"""
    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(None)
    """mapping of logical qubit names to physical qubit names, or None if using physical qubit names"""
    shots: int = Field(..., gt=0)
    """how many times to execute each circuit in the batch, must be greater than zero"""
    max_circuit_duration_over_t2: Optional[float] = Field(None)
    """Circuits are disqualified on the server if they are longer than this ratio
        of the T2 time of the qubits.
        If set to 0.0, no circuits are disqualified. If set to None the server default value is used."""
    heralding_mode: HeraldingMode = Field(HeraldingMode.NONE)
    """which heralding mode to use during the execution of circuits in this request."""

    move_validation_mode: MoveGateValidationMode = Field(MoveGateValidationMode.STRICT)
    """Which method of MOVE gate validation to use for circuit compilation."""
    move_gate_frame_tracking_mode: MoveGateFrameTrackingMode = Field(MoveGateFrameTrackingMode.FULL)
    """Which method of MOVE gate frame tracking to use for circuit compilation."""
    active_reset_cycles: Optional[int] = Field(None)
    """Number of active ``reset`` operations inserted at the beginning of each circuit for each active qubit.
    ``None`` means active reset is not used but instead reset is done by waiting (relaxation). Integer values smaller
    than 1 result in neither active nor reset by wait being used, in which case any reset operations must be explicitly
    added in the circuit."""
    dd_mode: DDMode = Field(DDMode.DISABLED)
    """Control whether dynamical decoupling should be enabled or disabled during the execution."""
    dd_strategy: Optional[DDStrategy] = Field(None)
    """A particular dynamical decoupling strategy to be used during the execution."""


CircuitMeasurementResults = dict[str, list[list[int]]]
"""Measurement results from a single circuit. For each measurement operation in the circuit,
maps the measurement key to the corresponding results. The outer list elements correspond to shots,
and the inner list elements to the qubits measured in the measurement operation."""

CircuitMeasurementResultsBatch = list[CircuitMeasurementResults]
"""Type that represents measurement results for a batch of circuits."""


class JobParameters(BaseModel):
    """Job-specific parameters extracted from the original RunRequest."""

    shots: int = Field(...)
    max_circuit_duration_over_t2: Optional[float] = Field(None)
    heralding_mode: HeraldingMode = Field(HeraldingMode.NONE)
    move_validation_mode: MoveGateValidationMode = Field(MoveGateValidationMode.STRICT)
    move_gate_frame_tracking_mode: MoveGateFrameTrackingMode = Field(MoveGateFrameTrackingMode.FULL)
    dd_mode: DDMode = Field(DDMode.DISABLED)
    dd_strategy: Optional[DDStrategy] = Field(None)


class Metadata(BaseModel):
    """Metadata describing a circuit execution job."""

    calibration_set_id: Optional[UUID] = Field(None)
    """ID of the calibration set used"""
    request: Optional[RunRequest] = Field(None)
    """optional copy of the original RunRequest sent to the server"""
    parameters: Optional[JobParameters] = Field(None)
    """job-specific parameters extracted from the original request"""
    circuits_batch: Optional[CircuitBatch] = Field(None)
    """circuits batch submitted for execution"""
    cocos_version: Optional[str] = Field(None)
    """CoCoS version used to execute the job"""
    timestamps: Optional[dict[str, str]] = Field(None)
    """Timestamps of execution progress"""

    @property
    def shots(self) -> int:
        """Return the number of shots in the job."""
        if self.parameters is not None:
            return self.parameters.shots
        if self.request is not None:
            return self.request.shots
        raise ValueError('No shots information available in the metadata')

    @property
    def circuits(self) -> CircuitBatch:
        """Return the circuits in the job."""
        if self.circuits_batch is not None:
            return self.circuits_batch
        if self.request is not None:
            return self.request.circuits
        raise ValueError('No circuits information available in the metadata')

    @property
    def heralding_mode(self) -> HeraldingMode:
        """Return the heralding mode requested with the job."""
        if self.parameters is not None:
            return self.parameters.heralding_mode
        if self.request is not None:
            return self.request.heralding_mode
        raise ValueError('No heralding mode information available in the metadata')

    @property
    def dd_mode(self) -> DDMode:
        """Return the dynamical decoupling mode requested with the job."""
        if self.parameters is not None:
            return self.parameters.dd_mode
        if self.request is not None:
            return self.request.dd_mode
        raise ValueError('No dynamical decoupling mode information available in the metadata')

    @property
    def dd_strategy(self) -> Optional[DDStrategy]:
        """Return the dynamical decoupling strategy used with the job."""
        if self.parameters is not None:
            return self.parameters.dd_strategy
        if self.request is not None:
            return self.request.dd_strategy
        raise ValueError('No dynamical decoupling strategy information available in the metadata')


class Status(str, Enum):
    """
    Status of a job.
    """

    RECEIVED = 'received'
    PROCESSING = 'processing'
    ACCEPTED = 'accepted'

    PENDING_COMPILATION = 'pending compilation'
    PENDING_EXECUTION = 'pending execution'
    READY = 'ready'
    FAILED = 'failed'
    ABORTED = 'aborted'
    PENDING_DELETION = 'pending deletion'
    DELETION_FAILED = 'deletion failed'
    DELETED = 'deleted'


class RunResult(BaseModel):
    """
    Results of the quantum circuit execution job.
    If the job succeeded, :attr:`measurements` contains the output of the batch of circuits,
    consisting of the results of the measurement operations in each circuit.
    It is a list of dictionaries, where each dict maps each measurement key to a 2D array of measurement
    results, represented as a nested list.
    ``RunResult.measurements[circuit_index][key][shot][qubit_index]`` is the result of measuring the
    ``qubit_index``'th qubit in measurement operation ``key`` in the shot ``shot`` in the
    ``circuit_index``'th circuit of the batch.
    :attr:`measurements` is present iff the status is ``'ready'``.
    :attr:`message` carries additional information for the ``'failed'`` status.
    If the status is ``'pending compilation'`` or ``'pending execution'``,
    :attr:`measurements` and :attr:`message` are ``None``.

    The results are non-negative integers representing the computational basis state (for qubits, 0 or 1)
    that was the measurement outcome.

    ----
    """

    status: Status = Field(...)
    """current status of the job, in ``{'pending compilation', 'pending execution', 'ready', 'failed', 'aborted'}``"""
    measurements: Optional[CircuitMeasurementResultsBatch] = Field(None)
    """if the job has finished successfully, the measurement results for the circuit(s)"""
    message: Optional[str] = Field(None)
    """if the job failed, an error message"""
    metadata: Metadata = Field(...)
    """metadata about the job"""
    warnings: Optional[list[str]] = Field(None)
    """list of warning messages"""

    @staticmethod
    def from_dict(inp: dict[str, Union[str, dict, list, None]]) -> RunResult:
        """Parses the result from a dict.

        Args:
            inp: value to parse, has to map to RunResult

        Returns:
            parsed job result

        """
        input_copy = inp.copy()
        return RunResult(status=Status(input_copy.pop('status')), **input_copy)


class RunStatus(BaseModel):
    """Status of a circuit execution job."""

    status: Status = Field(...)
    """current status of the job, in ``{'pending compilation', 'pending execution', 'ready', 'failed', 'aborted'}``"""
    message: Optional[str] = Field(None)
    """if the job failed, an error message"""
    warnings: Optional[list[str]] = Field(None)
    """list of warning messages"""

    @staticmethod
    def from_dict(inp: dict[str, Union[str, dict, list, None]]) -> RunStatus:
        """Parses the result from a dict.

        Args:
            inp: value to parse, has to map to RunResult

        Returns:
            parsed job status

        """
        input_copy = inp.copy()
        return RunStatus(status=Status(input_copy.pop('status')), **input_copy)


class Counts(BaseModel):
    """Measurement results in the counts representation"""

    measurement_keys: list[str]
    """measurement keys in the order they are concatenated to form the states in counts"""
    counts: dict[str, int]
    """counts as a dictionary mapping states represented as bitstrings to the number of shots they were measured"""


class RunCounts(BaseModel):
    """Measurement counts of a circuit execution job."""

    status: Status = Field(...)
    """current status of the job, in ``{'pending compilation', 'pending execution', 'ready', 'failed', 'aborted'}``"""
    counts_batch: Optional[list[Counts]] = Field(
        None,
        description="""Measurement results in histogram representation.
    The `measurement_keys` list provides the order of the measurment keys for the repesentation of the states in the keys
    of the `counts` dictionary. As an example if `measurement_keys` is `['mk_1', 'mk2']` and `mk_1` refers to `QB1` and `mk_2`
    refers to `QB3` and `QB5` then counts could contains keys such as '010' with `QB1` in the 0, `QB3` in the 1 and `QB5` in
    the 0 state.""",
    )

    @staticmethod
    def from_dict(inp: dict[str, Union[str, dict, list, None]]) -> RunCounts:
        """Parses the result from a dict.

        Args:
            inp: value to parse, has to map to RunCounts

        Returns:
            parsed job status

        """
        input_copy = inp.copy()
        return RunCounts(status=Status(input_copy.pop('status')), **input_copy)


class ClientLibrary(BaseModel):
    """Represents a client library with its metadata.

    Args:
        name: display name of the client library.
        package_name: name of the package as published in package repositories.
        repo_url: URL to the source code repository.
        package_url: URL to the package in the package repository.
        min: minimum supported version.
        max: maximum supported version.
    """

    name: str
    package_name: Optional[str] = Field(None)
    repo_url: Optional[str] = Field(None)
    package_url: Optional[str] = Field(None)
    min: str
    max: str



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains definitions of IQM Server API endpoints."""

from enum import Enum, auto
from posixpath import join


class APIEndpoint(Enum):
    """Supported API endpoints."""

    CONFIGURATION = auto()
    SUBMIT_JOB = auto()
    GET_JOB_REQUEST_PARAMETERS = auto()
    GET_JOB_RESULT = auto()
    GET_JOB_STATUS = auto()
    GET_JOB_CALIBRATION_SET_ID = auto()
    GET_JOB_CIRCUITS_BATCH = auto()
    GET_JOB_TIMELINE = auto()
    GET_JOB_ERROR_LOG = auto()
    GET_JOB_COUNTS = auto()
    ABORT_JOB = auto()
    DELETE_JOB = auto()
    HEALTH = auto()
    ABOUT = auto()
    CLIENT_LIBRARIES = auto()

    # Calibration and Calibration Service endpoints
    CALIBRATION_SERVICE_CONFIGURATION = auto()
    QUANTUM_ARCHITECTURE = auto()
    QUALITY_METRICS_LATEST = auto()
    QUALITY_METRICS_MONITORING = auto()
    CALIBRATED_GATES = auto()
    START_CALIBRATION_JOB = auto()
    ABORT_CALIBRATION_JOB = auto()
    CALIBRATION_SERVICE_JOBS = auto()
    CALIBRATION = auto()


class APIVariant(Enum):
    """
    Supported API versions and variants.

    WARNING:
    Only .V1 is considered stable. The .V2 API is experimental and
    is subject to change without notice, even in minor or patch releases.
    Use with caution, as future updates may introduce breaking changes or remove functionality entirely.
    """

    V1 = "V1"  # QCCSW Cocos-based circuits execution
    V2 = "V2"  # Station-Control-based circuits execution
    RESONANCE_COCOS_V1 = "RESONANCE_COCOS_V1"  # IQM Resonance Cocos API


class APIConfig:
    """Provides supported API endpoints for a given API variant."""

    def __init__(self, variant: APIVariant, iqm_server_url: str):
        """
        Args:
            variant: API variant.
            iqm_server_url: URL of the IQM server,
                            e.g. https://test.qc.iqm.fi/cocos or https://cocos.resonance.meetiqm.com/garnet for .V1
                            or https://test.qc.iqm.fi for .V2
        """
        self.variant = variant
        self.iqm_server_url = iqm_server_url
        self.urls = self._get_api_urls()

    def _get_api_urls(self) -> dict[APIEndpoint, str]:
        """
        Returns:
            Relative URLs for each supported API endpoints.
        """
        if self.variant == APIVariant.RESONANCE_COCOS_V1:
            return {
                APIEndpoint.SUBMIT_JOB: "jobs",
                APIEndpoint.GET_JOB_RESULT: "jobs/%s",
                APIEndpoint.GET_JOB_STATUS: "jobs/%s/status",
                APIEndpoint.GET_JOB_COUNTS: "jobs/%s/counts",
                APIEndpoint.ABORT_JOB: "jobs/%s/abort",
                APIEndpoint.QUANTUM_ARCHITECTURE: "quantum-architecture",
                APIEndpoint.CALIBRATED_GATES: "api/v1/calibration/%s/gates",
                APIEndpoint.CLIENT_LIBRARIES: "info/client-libraries",
            }
        if self.variant == APIVariant.V1:
            return {
                APIEndpoint.CONFIGURATION: "configuration",
                APIEndpoint.QUALITY_METRICS_LATEST: "calibration/metrics/latest",
                APIEndpoint.SUBMIT_JOB: "jobs",
                APIEndpoint.GET_JOB_RESULT: "jobs/%s",
                APIEndpoint.GET_JOB_STATUS: "jobs/%s/status",
                APIEndpoint.GET_JOB_COUNTS: "jobs/%s/counts",
                APIEndpoint.ABORT_JOB: "jobs/%s/abort",
                APIEndpoint.ABORT_CALIBRATION_JOB: "jobs/%s/abort",
                APIEndpoint.DELETE_JOB: "jobs/%s",
                APIEndpoint.QUANTUM_ARCHITECTURE: "quantum-architecture",
                APIEndpoint.CALIBRATED_GATES: "api/v1/calibration/%s/gates",
                APIEndpoint.QUALITY_METRICS_MONITORING: "api/v1/monitor/calibration/metrics",
                APIEndpoint.HEALTH: "health",
                APIEndpoint.ABOUT: "about",
                APIEndpoint.CLIENT_LIBRARIES: "info/client-libraries",
                APIEndpoint.START_CALIBRATION_JOB: "calibration/run",
                APIEndpoint.CALIBRATION_SERVICE_CONFIGURATION: "calibration/configuration",
                APIEndpoint.CALIBRATION_SERVICE_JOBS: "calibration/jobs",
                APIEndpoint.CALIBRATION: "api/v1/calibration/%s",
            }
        if self.variant == APIVariant.V2:
            return {
                APIEndpoint.GET_JOB_REQUEST_PARAMETERS: "station/circuits/%s/request_parameters",
                APIEndpoint.CONFIGURATION: "cocos/configuration",
                APIEndpoint.QUALITY_METRICS_LATEST: "cocos/calibration/metrics/latest",
                APIEndpoint.SUBMIT_JOB: "station/circuits",
                APIEndpoint.GET_JOB_RESULT: "station/circuits/%s/measurements",
                APIEndpoint.GET_JOB_STATUS: "station/circuits/%s/status",
                APIEndpoint.GET_JOB_CALIBRATION_SET_ID: "station/circuits/%s/calibration_set_id",
                APIEndpoint.GET_JOB_CIRCUITS_BATCH: "station/circuits/%s/circuits_batch",
                APIEndpoint.GET_JOB_TIMELINE: "station/circuits/%s/timeline",
                APIEndpoint.GET_JOB_ERROR_LOG: "station/circuits/%s/error_log",
                APIEndpoint.GET_JOB_COUNTS: "station/circuits/%s/counts",
                APIEndpoint.ABORT_JOB: "station/circuits/%s/abort",
                APIEndpoint.ABORT_CALIBRATION_JOB: "cocos/jobs/%s/abort",
                APIEndpoint.DELETE_JOB: "station/circuits/%s",
                APIEndpoint.QUANTUM_ARCHITECTURE: "cocos/quantum-architecture",
                APIEndpoint.CALIBRATED_GATES: "cocos/api/v1/calibration/%s/gates",
                APIEndpoint.QUALITY_METRICS_MONITORING: "cocos/api/v1/monitor/calibration/metrics",
                APIEndpoint.HEALTH: "cocos/health",
                APIEndpoint.ABOUT: "cocos/about",
                APIEndpoint.CLIENT_LIBRARIES: "info/client-libraries",
                APIEndpoint.START_CALIBRATION_JOB: "cocos/calibration/run",
                APIEndpoint.CALIBRATION_SERVICE_CONFIGURATION: "cocos/calibration/configuration",
                APIEndpoint.CALIBRATION_SERVICE_JOBS: "cocos/calibration/jobs",
                APIEndpoint.CALIBRATION: "cocos/api/v1/calibration/%s",
            }
        raise ValueError(f"Unsupported API variant: {self.variant}")

    def is_supported(self, endpoint: APIEndpoint) -> bool:
        """
        Args:
            endpoint: API endpoint.

        Returns:
            True if the endpoint is supported, False otherwise.
        """
        return endpoint in self.urls

    def url(self, endpoint: APIEndpoint, *args) -> str:
        """
        Args:
            endpoint: API endpoint.
            args: Arguments to be passed to the URL.

        Returns:
            URL for the given endpoint.

        Raises:
            ValueError: If the endpoint is not supported.
        """
        url = self.urls.get(endpoint)
        if url is None:
            raise ValueError(f"Unsupported API endpoint: {endpoint}")
        return join(self.iqm_server_url, url % args)



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the IQM client utilities.
"""
import numpy as np
import pytest

from iqm.iqm_client.util import to_json_dict


def test_serialize_dict():
    """
    Tests that util.to_json_dict can handle a dict that is already JSON compatible
    """
    data = {"key1": True, "key2": None, "key3": 123.456, "key4": {"key5": ["a", "b", "c"], "key6": "value6"}}
    assert to_json_dict(data) == data


def test_serialize_dict_with_ndarray():
    """
    Tests that util.to_json_dict can handle a dict that contains numpy.ndarray value
    """
    original = np.array([[1, 2, 3], [4, 5, 6]])
    expected = [[1, 2, 3], [4, 5, 6]]
    json_dict = to_json_dict({"key1": {"key2": original}})
    assert json_dict == {"key1": {"key2": expected}}


def test_serialize_dict_with_unsupported_value():
    """
    Tests that util.to_json_dict raises ValueError if there is unsupported data in the dict.

    `to_json_dict` catches TypeError raised from JSON serialization and raises a ValueError from it.
    """
    original = {"key1": complex(1.0, 2.0)}
    with pytest.raises(ValueError):
        to_json_dict(original)


def test_serialize_dict_with_nan_value():
    """
    Tests that util.to_json_dict raises TypeError if there is unsupported data in the dict

    `to_json_dict` catches ValueError raised from JSON serialization and raises a new ValueError from it.
    """
    original = {"key1": float("NaN")}
    with pytest.raises(ValueError):
        to_json_dict(original)



# Copyright 2021-2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Collection of transpilation functions needed for transpiling to specific devices.
"""
from enum import Enum
from typing import Any, Iterable, Optional
import warnings

from iqm.iqm_client import (
    Circuit,
    CircuitTranspilationError,
    CircuitValidationError,
    DynamicQuantumArchitecture,
    Instruction,
    IQMClient,
)


class ExistingMoveHandlingOptions(str, Enum):
    """Transpile options for handling of existing MOVE instructions."""

    KEEP = 'keep'
    """Transpiler will keep the MOVE instructions as specified checking if they are correct, adding more as needed."""
    REMOVE = 'remove'
    """Transpiler will remove the instructions and add new ones."""
    TRUST = 'trust'
    """Transpiler will keep the MOVE instructions without checking if they are correct, and add more as needed."""


class ResonatorStateTracker:
    r"""Class for tracking the location of the :math:`|0\rangle` state of the resonators on the
    quantum computer as they are moved with the MOVE gates because the MOVE gate is not defined
    when acting on a :math:`|11\rangle` state. This is equivalent to tracking
    which qubit state has been moved into which resonator.

    Args:
        available_moves: A dictionary describing between which qubits a MOVE gate is
            available, for each resonator, i.e. ``available_moves[resonator] = [qubit]``
    """

    move_gate = 'move'

    def __init__(self, available_moves: dict[str, list[str]]) -> None:
        self.available_moves = available_moves
        self.res_qb_map = {r: r for r in self.resonators}

    @staticmethod
    def from_dynamic_architecture(arch: DynamicQuantumArchitecture) -> 'ResonatorStateTracker':
        """Constructor to make the ResonatorStateTracker from a DynamicQuantumArchitecture.

        Args:
            arch: The architecture to track the resonator state on.
        """
        available_moves: dict[str, list[str]] = {
            r: [q for q, r2 in arch.gates[ResonatorStateTracker.move_gate].loci if r == r2]
            for r in arch.computational_resonators
        }
        return ResonatorStateTracker(available_moves)

    @staticmethod
    def from_circuit(circuit: Circuit) -> 'ResonatorStateTracker':
        """Constructor to make the ResonatorStateTracker from a circuit.

        Infers the resonator connectivity from the MOVE gates in the circuit.

        Args:
            circuit: The circuit to track the resonator state on.
        """
        return ResonatorStateTracker.from_instructions(circuit.instructions)

    @staticmethod
    def from_instructions(instructions: Iterable[Instruction]) -> 'ResonatorStateTracker':
        """Constructor to make the ResonatorStateTracker from a sequence of instructions.

        Infers the resonator connectivity from the MOVE gates.

        Args:
            instructions: The instructions to track the resonator state on.
        """
        available_moves: dict[str, list[str]] = {}
        for i in instructions:
            if i.name == ResonatorStateTracker.move_gate:
                q, r = i.qubits
                available_moves.setdefault(r, []).append(q)
        return ResonatorStateTracker(available_moves)

    @property
    def resonators(self) -> Iterable[str]:
        """Getter for the resonator registers that are being tracked."""
        return self.available_moves

    @property
    def supports_move(self) -> bool:
        """Bool whether any MOVE gate is allowed."""
        return bool(self.available_moves)

    def apply_move(self, qubit: str, resonator: str) -> None:
        """Apply the logical changes of the resonator state location when a MOVE gate between qubit and resonator is
        applied.

        Args:
            qubit: The moved qubit.
            resonator: The moved resonator.

        Raises:
            CircuitTranspilationError: MOVE is not allowed, either because the resonator does not exist,
                the MOVE gate is not available between this qubit-resonator pair, or the resonator state
                is currently in a different qubit register.
        """
        if (
            resonator in self.resonators
            and qubit in self.available_moves[resonator]
            and self.res_qb_map[resonator] in [qubit, resonator]
        ):
            self.res_qb_map[resonator] = qubit if self.res_qb_map[resonator] == resonator else resonator
        else:
            raise CircuitTranspilationError('Attempted move is not allowed.')

    def create_move_instructions(
        self,
        qubit: str,
        resonator: str,
        apply_move: Optional[bool] = True,
        alt_qubit_names: Optional[dict[str, str]] = None,
    ) -> Iterable[Instruction]:
        """Create the MOVE instructions needed to move the given resonator state into the resonator if needed and then
        move resonator state to the given qubit.

        Args:
            qubit: The qubit
            resonator: The resonator
            apply_move: Whether the moves should be applied to the resonator tracking state.
            alt_qubit_names: Mapping of logical qubit names to physical qubit names.

        Yields:
            The one or two MOVE instructions needed.
        """
        if self.res_qb_map[resonator] not in [qubit, resonator]:
            other = self.res_qb_map[resonator]
            if apply_move:
                self.apply_move(other, resonator)
            qbs = tuple(alt_qubit_names[q] if alt_qubit_names else q for q in [other, resonator])
            yield Instruction(name=self.move_gate, qubits=qbs, args={})
        if apply_move:
            self.apply_move(qubit, resonator)
        qbs = tuple(alt_qubit_names[q] if alt_qubit_names else q for q in [qubit, resonator])
        yield Instruction(name=self.move_gate, qubits=qbs, args={})

    def reset_as_move_instructions(
        self,
        resonators: Optional[Iterable[str]] = None,
        apply_move: Optional[bool] = True,
        alt_qubit_names: Optional[dict[str, str]] = None,
    ) -> list[Instruction]:
        """Creates the MOVE instructions needed to move all resonator states to their original state.

        Args:
            resonators: The set of resonators to reset, if None, all resonators will be reset.
            apply_move: Whether the moves should be applied to the resonator tracking state.

        Returns:
            The instructions needed to move all qubit states out of the resonators.
        """
        if resonators is None:
            resonators = self.resonators
        instructions: list[Instruction] = []
        for r, q in [(r, q) for r, q in self.res_qb_map.items() if r != q and r in resonators]:
            instructions += self.create_move_instructions(q, r, apply_move, alt_qubit_names)
        return instructions

    def available_resonators_to_move(self, qubits: Iterable[str]) -> dict[str, list[str]]:
        """Generates a dictionary with which resonators a qubit can be moved to, for each qubit.

        Args:
            qubits: The qubits to check which MOVE gates are available.

        Returns:
            The dict that maps each qubit to a list of resonators.
        """
        return {q: [r for r in self.resonators if q in self.available_moves[r]] for q in qubits}

    def resonators_holding_qubits(self, qubits: Iterable[str]) -> list[str]:
        """Returns the resonators that are currently holding one of the given qubit states.

        Args:
            qubits: The qubits

        Returns:
            The resonators
        """
        return [r for r, q in self.res_qb_map.items() if q in qubits and q not in self.resonators]

    def choose_move_pair(
        self, qubits: list[str], remaining_instructions: list[list[str]]
    ) -> list[tuple[str, str, list[list[str]]]]:
        """Chooses which qubit of the given qubits to move into which resonator, given a sequence of instructions to be
        executed later for looking ahead.

        Args:
            qubits: The qubits to choose from
            remaining_instructions: The instructions to use for the look-ahead.

        Raises:
            CircuitTranspilationError: When no move pair is available, most likely because the circuit was not routed.

        Returns:
            A sorted preference list of resonator and qubit chosen to apply the move on.
        """
        r_candidates = [
            (r, q, remaining_instructions) for q, rs in self.available_resonators_to_move(qubits).items() for r in rs
        ]
        if len(r_candidates) == 0:
            raise CircuitTranspilationError(
                f'Unable to insert MOVE gates because none of the qubits {qubits} share a resonator. '
                + 'This can be resolved by routing the circuit first without resonators.'
            )
        resonator_candidates = list(sorted(r_candidates, key=self._score_choice_heuristic, reverse=True))
        return resonator_candidates

    def _score_choice_heuristic(self, args: tuple[str, str, list[list[str]]]) -> int:
        """A simple look ahead heuristic for choosing which qubit to move where.

        Counts the number of CZ gates until the qubit needs to be moved out.

        Args:
            args: resonator, qubit, instructions.

        Returns:
            The count/score.
        """
        _, qb, circ = args
        score: int = 0
        for instr in circ:
            if qb in instr:
                if instr[0] != 'cz':
                    return score
                score += 1
        return score

    def update_qubits_in_resonator(self, qubits: Iterable[str]) -> list[str]:
        """Applies the resonator to qubit map in the state of the resonator state tracker to the given qubits.

        Args:
            qubits: The qubits or resonators to apply the state to

        Returns:
            The remapped qubits
        """
        return [self.res_qb_map.get(q, q) for q in qubits]


def transpile_insert_moves(
    circuit: Circuit,
    arch: DynamicQuantumArchitecture,
    existing_moves: Optional[ExistingMoveHandlingOptions] = None,
    qubit_mapping: Optional[dict[str, str]] = None,
) -> Circuit:
    """Inserts MOVEs to the circuit according to a given architecture specification.

    The function does nothing if the given architecture specification does not support MOVE gates.
    Note that this method assumes that the circuit is already transpiled to a coupling map/architecture where the
    resonator has been abstracted away, i.e. the edges of the coupling map that contain resonators are replaced by
    edges between the other qubit and all qubits that can be moved to that resonator.

    Args:
        circuit: The circuit to add MOVE instructions to.
        arch: Restrictions of the target device
        existing_moves: Specifies how to deal with existing MOVE instruction,
            If None, the function will use ExistingMoveHandlingOptions.REMOVE with a user warning if there are move
            instructions in the circuit.
        qubit_mapping: Mapping of logical qubit names to physical qubit names.
            Can be set to ``None`` if all ``circuits`` already use physical qubit names.
    """
    res_status = ResonatorStateTracker.from_dynamic_architecture(arch)
    if not qubit_mapping:
        qubit_mapping = {}
    for q in arch.components:
        if q not in qubit_mapping.values():
            qubit_mapping[q] = q
    existing_moves_in_circuit = [i for i in circuit.instructions if i.name == res_status.move_gate]

    if existing_moves is None and len(existing_moves_in_circuit) > 0:
        warnings.warn('Circuit already contains MOVE instructions, removing them before transpiling.')
        existing_moves = ExistingMoveHandlingOptions.REMOVE

    if not res_status.supports_move:
        if not existing_moves_in_circuit:
            return circuit
        if existing_moves == ExistingMoveHandlingOptions.REMOVE:
            return transpile_remove_moves(circuit)
        raise ValueError('Circuit contains MOVE instructions, but device does not support them')

    if existing_moves is None or existing_moves == ExistingMoveHandlingOptions.REMOVE:
        circuit = transpile_remove_moves(circuit)
    elif existing_moves == ExistingMoveHandlingOptions.KEEP:
        try:
            IQMClient._validate_circuit_moves(arch, circuit, qubit_mapping=qubit_mapping)
        except CircuitValidationError as e:
            raise CircuitTranspilationError(
                f'Unable to transpile the circuit after validation error: {e.args[0]}'
            ) from e

    rev_qubit_mapping = {v: k for k, v in qubit_mapping.items()}
    new_instructions = _transpile_insert_moves(list(circuit.instructions), res_status, arch, qubit_mapping)
    new_instructions += res_status.reset_as_move_instructions(alt_qubit_names=rev_qubit_mapping)

    return Circuit(name=circuit.name, instructions=new_instructions, metadata=circuit.metadata)


def _transpile_insert_moves(
    instructions: list[Instruction],
    res_status: ResonatorStateTracker,
    arch: DynamicQuantumArchitecture,
    qubit_mapping: dict[str, str],
) -> list[Instruction]:
    """Inserts MOVE gates into a list of instructions and changes the existing instructions as needed.

    Helper function for :func:`transpile_insert_moves`.

    Args:
        instructions: The instructions in the circuit.
        res_status: The location of the resonator states at the start of the instructions. At
        the end of this method this tracker is adjusted to reflect the state at the end of the returned instructions.
        arch: The target quantum architecture.
        qubit_mapping: Mapping from logical qubit names to physical qubit names.

    Raises:
        CircuitTranspilationError: Raised when the circuit contains invalid gates that cannot be transpiled using this
            method.

    Returns:
        The transpiled list of instructions.
    """
    new_instructions = []
    rev_qubit_mapping = {v: k for k, v in qubit_mapping.items()}
    for idx, i in enumerate(instructions):
        qubits = [qubit_mapping[q] for q in i.qubits]
        res_match = res_status.resonators_holding_qubits(qubits)
        if res_match and i.name not in ['cz', res_status.move_gate]:
            # We have a gate on a qubit in the resonator that cannot be executed on the resonator (incl. barriers)
            new_instructions += res_status.reset_as_move_instructions(res_match, alt_qubit_names=rev_qubit_mapping)
            new_instructions.append(i)
        else:
            # Check if the instruction is valid, which raises an exception if not.
            try:
                IQMClient._validate_instruction(
                    architecture=arch,
                    instruction=i,
                    qubit_mapping=qubit_mapping,
                )
                new_instructions.append(i)  # No adjustment needed
                if i.name == res_status.move_gate:  # update the tracker if needed
                    res_status.apply_move(*[qubit_mapping[q] for q in i.qubits])
            except CircuitValidationError as e:
                if i.name != 'cz':  # We can only fix cz gates at this point
                    raise CircuitTranspilationError(
                        f'Unable to transpile the circuit after validation error: {e.args[0]}'
                    ) from e
                # Pick which qubit-resonator pair to apply this cz to
                # Pick from qubits already in a resonator or both targets if none off them are in a resonator
                resonator_candidates: Optional[list[tuple[str, str, Any]]] = res_status.choose_move_pair(
                    [res_status.res_qb_map[res] for res in res_match] if res_match else qubits,
                    [[i.name] + [qubit_mapping[q] for q in i.qubits] for i in instructions[idx:]],
                )
                while resonator_candidates:
                    r, q1, _ = resonator_candidates.pop(0)
                    q2 = [q for q in qubits if q != q1][0]
                    try:
                        IQMClient._validate_instruction(
                            architecture=arch,
                            instruction=Instruction(
                                name='cz', qubits=(rev_qubit_mapping[q2], rev_qubit_mapping[r]), args={}
                            ),
                            qubit_mapping=qubit_mapping,
                        )
                        resonator_candidates = None
                        break
                    except CircuitValidationError:
                        pass

                if resonator_candidates is not None:
                    raise CircuitTranspilationError(
                        'Unable to find a valid resonator-qubit pair for a MOVE gate to enable this CZ gate.'
                    ) from e

                # remove the other qubit from the resonator if it was in
                new_instructions += res_status.reset_as_move_instructions(
                    [res for res in res_match if res != r], alt_qubit_names=rev_qubit_mapping
                )
                # move the qubit into the resonator if it was not yet in.
                if not res_match:
                    new_instructions += res_status.create_move_instructions(q1, r, alt_qubit_names=rev_qubit_mapping)
                new_instructions.append(
                    Instruction(name='cz', qubits=(rev_qubit_mapping[q2], rev_qubit_mapping[r]), args={})
                )
    return new_instructions


def transpile_remove_moves(circuit: Circuit) -> Circuit:
    """Removes MOVE gates from a circuit.

    The method assumes that these MOVE gates are moving the resonator state in and out the resonator register to
    reconstruct the CZ gates. If this is not the case, the semantic equivalence cannot be guaranteed.

    Args:
        circuit: The circuit from which the MOVE gates need to be removed.

    Returns:
        The circuit with the MOVE gates removed and the targets for all other gates updated accordingly.
    """
    res_status = ResonatorStateTracker.from_circuit(circuit)
    new_instructions = []
    for i in circuit.instructions:
        if i.name == res_status.move_gate:
            res_status.apply_move(*i.qubits)
        else:
            new_qubits = res_status.update_qubits_in_resonator(i.qubits)
            new_instructions.append(Instruction(name=i.name, qubits=new_qubits, args=i.args))
    return Circuit(name=circuit.name, instructions=new_instructions, metadata=circuit.metadata)



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from iqm.iqm_client import QuantumArchitectureSpecification


@pytest.mark.parametrize(
    'operations',
    [
        {'prx': [['QB1'], ['QB2'], ['QB3']]},
        {'prx': [['QB2'], ['QB1'], ['QB3']]},
        {'prx': [['QB3'], ['QB1'], ['QB2']]},
    ],
)
def test_equivalent_prx_ops(operations):
    verify_ops_match({'prx': [['QB1'], ['QB2'], ['QB3']]}, operations, True)


@pytest.mark.parametrize(
    'operations',
    [
        {'prx': [['QB2'], ['QB2'], ['QB3']]},
        {'prx': [['QB2'], ['QB3']]},
        {'prx': [['QB3'], ['QB1'], ['QB2'], ['QB4']]},
        {'prx': [['QB1'], ['QB2'], ['QB3']], 'move': [['QB1', 'QB2']]},
    ],
)
def test_different_prx_ops(operations):
    verify_ops_match({'prx': [['QB1'], ['QB2'], ['QB3']]}, operations, False)


@pytest.mark.parametrize(
    'operations',
    [
        {'prx': [['QB1'], ['QB2'], ['QB3']], 'move': [['QB1', 'QB2']]},
        {'prx': [['QB2'], ['QB1'], ['QB3']], 'move': [['QB1', 'QB2']]},
        {'prx': [['QB3'], ['QB1'], ['QB2']], 'move': [['QB1', 'QB2']]},
    ],
)
def test_equivalent_prx_move_ops(operations):
    verify_ops_match({'prx': [['QB1'], ['QB2'], ['QB3']], 'move': [['QB1', 'QB2']]}, operations, True)


@pytest.mark.parametrize(
    'operations',
    [
        {'prx': [['QB1'], ['QB2'], ['QB3']], 'move': [['QB2', 'QB1']]},
        {'prx': [['QB2'], ['QB1']], 'move': [['QB1', 'QB2']]},
        {'prx': [['QB3'], ['QB1'], ['QB2']], 'move': [['QB2', 'QB3']]},
        {'prx': [['QB3'], ['QB1'], ['QB2']], 'move': [['QB1', 'QB2'], ['QB2', 'QB3']]},
    ],
)
def test_different_prx_move_ops(operations):
    verify_ops_match({'prx': [['QB1'], ['QB2'], ['QB3']], 'move': [['QB1', 'QB2']]}, operations, False)


@pytest.mark.parametrize(
    'operations',
    [
        {'prx': [['QB1'], ['QB2'], ['QB3']], 'cz': [['QB2', 'QB1']]},
        {'prx': [['QB2'], ['QB1'], ['QB3']], 'cz': [['QB2', 'QB1']]},
        {'prx': [['QB3'], ['QB1'], ['QB2']], 'cz': [['QB1', 'QB2']]},
    ],
)
def test_equivalent_prx_cz_ops(operations):
    verify_ops_match({'prx': [['QB1'], ['QB2'], ['QB3']], 'cz': [['QB1', 'QB2']]}, operations, True)


def verify_ops_match(ops1, ops2, should_match: bool):
    assert QuantumArchitectureSpecification.compare_operations(ops1=ops1, ops2=ops2) == should_match
    assert QuantumArchitectureSpecification.compare_operations(ops1=ops2, ops2=ops1) == should_match
    arch1 = to_arch(ops1)
    arch2 = to_arch(ops2)
    assert arch1.has_equivalent_operations(arch2) == should_match
    assert arch2.has_equivalent_operations(arch1) == should_match


def to_arch(operations):
    return QuantumArchitectureSpecification(
        name='hercules',
        qubits=['QB1', 'QB2', 'QB3'],
        qubit_connectivity=[['QB1', 'QB2'], ['QB2', 'QB3']],
        operations=operations,
    )


def test_simplified_quantum_architecture():
    simplified_architecture = QuantumArchitectureSpecification(
        name='hercules',
        qubits=['QB1', 'QB2', 'QB3'],
        qubit_connectivity=[['QB1', 'QB2'], ['QB2', 'QB3']],
        operations=['prx', 'cz'],
    )
    assert simplified_architecture.operations == {
        'prx': [['QB1'], ['QB2'], ['QB3']],
        'cz': [['QB1', 'QB2'], ['QB2', 'QB3']],
    }



import subprocess
from packaging.requirements import Requirement
import tomllib

if __name__ == "__main__":
    print("Parsing pyproject.toml for all direct project dependencies and optional dependencies.")
    with open("pyproject.toml", "rb") as file:
        pyproject = tomllib.load(file)

    dependencies = pyproject["project"]["dependencies"]
    dependencies += pyproject["project"]["optional-dependencies"]["testing"]
    dependencies += pyproject["project"]["optional-dependencies"]["docs"]
    dependencies += pyproject["project"]["optional-dependencies"]["cicd"]

    # NOTE: see comment about build deps in pyproject.toml.

    update_lines = [f"--upgrade-package={Requirement(dep).name}" for dep in dependencies]
    cmd = [
        "uv",
        "pip",
        "compile",
        "--generate-hashes",
        "--no-annotate",
        "--no-emit-index-url",
        "--no-emit-find-links",
        "--no-header",
        "--no-cache",
        "--python-platform=x86_64-manylinux_2_31",
        "--all-extras",
        "pyproject.toml",
        "--output-file=requirements.txt",
    ] + update_lines
    print(f"Executing `{subprocess.list2cmdline(cmd)}`")
    subprocess.run(cmd, check=True)



# Copyright 2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for user authentication and token management in IQM client.
"""
# pylint: disable=too-many-arguments
import json
from uuid import UUID, uuid4

from freezegun import freeze_time
from mockito import expect, mock, unstub, verifyNoUnwantedInteractions, when
from mockito.matchers import ANY
import pytest
import requests

from iqm.iqm_client import (
    REQUESTS_TIMEOUT,
    ClientAuthenticationError,
    ClientConfigurationError,
    IQMClient,
    RunResult,
    RunStatus,
)
from iqm.iqm_client.authentication import (
    AUTH_CLIENT_ID,
    AUTH_REALM,
    AUTH_REQUESTS_TIMEOUT,
    ExternalToken,
    TokenClient,
    TokenManager,
    TokenProviderInterface,
    TokensFileReader,
)
from tests.conftest import MockJsonResponse, make_token, mock_supported_client_libraries_response


@pytest.fixture()
def auth_server_url() -> str:
    """Authentication server base URL"""
    return 'https://example.com/auth'


@pytest.fixture()
def auth_realm() -> str:
    """Authentication realm for token client tests"""
    return 'test_realm'


@pytest.fixture()
def auth_username() -> str:
    """Username for getting token from an authentication server"""
    return 'some-user'


@pytest.fixture()
def auth_password() -> str:
    """Username for getting token from an authentication server"""
    return 'very-secret'


def _make_client_with_token(base_url) -> tuple[str, IQMClient]:
    when(requests).get(f'{base_url}/info/client-libraries', headers=ANY, timeout=ANY).thenReturn(
        mock_supported_client_libraries_response()
    )
    token = make_token('Bearer', 300)
    client = IQMClient(base_url, token=token)
    return token, client


def test_external_token_provides_token():
    """Tests that ExternalToken provides the configured token"""
    token = make_token('Bearer', 300)

    provider = ExternalToken(token)
    assert provider.get_token() == token
    with pytest.raises(ClientAuthenticationError, match='Can not close externally managed auth session'):
        provider.close()
    with pytest.raises(ClientAuthenticationError, match='No external token available'):
        provider.get_token()

    verifyNoUnwantedInteractions()
    unstub()


def test_tokens_file_reader_provides_token(tmp_path):
    """Tests that TokensFileReader provides the access token stored in the file"""
    path = str(tmp_path / 'tokens_file.json')
    token = make_token('Bearer', 300)
    with open(path, 'w', encoding='utf-8') as tokens_file:
        tokens_file.write(json.dumps({'access_token': token}))

    provider = TokensFileReader(path)
    assert provider.get_token() == token
    with pytest.raises(ClientAuthenticationError, match='Can not close externally managed auth session'):
        provider.close()
    with pytest.raises(ClientAuthenticationError, match='No tokens file available'):
        provider.get_token()

    verifyNoUnwantedInteractions()
    unstub()


def test_tokens_file_reader_file_not_found(tmp_path):
    """Tests that TokensFileReader raises ClientAuthenticationError if the configured file is not found"""
    path = str(tmp_path / 'tokens_file.json')
    provider = TokensFileReader(path)
    with pytest.raises(ClientAuthenticationError, match='Failed to read access token'):
        provider.get_token()

    verifyNoUnwantedInteractions()
    unstub()


def test_tokens_file_reader_file_contains_invalid_data(tmp_path):
    """Tests that TokensFileReader raises ClientAuthenticaitonError if the configured file contains invalid data"""
    path = str(tmp_path / 'tokens_file.json')
    with open(path, 'w', encoding='utf-8') as tokens_file:
        tokens_file.write('some-invalid-data')

    provider = TokensFileReader(path)
    with pytest.raises(ClientAuthenticationError, match='Failed to read access token'):
        provider.get_token()
    with pytest.raises(ClientAuthenticationError, match='Can not close externally managed auth session'):
        provider.close()
    with pytest.raises(ClientAuthenticationError, match='No tokens file available'):
        provider.get_token()

    verifyNoUnwantedInteractions()
    unstub()


def test_token_client_provides_token(auth_server_url, auth_realm, auth_username, auth_password):
    """Tests that TokenClient provides the access token acquired from an authentication server"""
    # pylint: disable=too-many-locals

    token_url = f'{auth_server_url}/realms/{auth_realm}/protocol/openid-connect/token'
    logout_url = f'{auth_server_url}/realms/{auth_realm}/protocol/openid-connect/logout'
    provider = TokenClient(auth_server_url, auth_realm, auth_username, auth_password)

    # login
    with freeze_time('2024-05-20 12:00:00'):
        login_data1 = {
            'client_id': AUTH_CLIENT_ID,
            'grant_type': 'password',
            'username': auth_username,
            'password': auth_password,
        }
        access_token1 = make_token('Bearer', 300)
        refresh_token1 = make_token('Refresh', 3000)

        expect(requests, times=1).post(token_url, data=login_data1, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
            MockJsonResponse(200, {'access_token': access_token1, 'refresh_token': refresh_token1})
        )

        assert provider.get_token() == access_token1

    # 1st refresh
    with freeze_time('2024-05-20 12:01:00'):
        refresh_data1 = {'client_id': AUTH_CLIENT_ID, 'grant_type': 'refresh_token', 'refresh_token': refresh_token1}
        access_token2 = make_token('Bearer', 300)
        refresh_token2 = make_token('Refresh', 3000)

        expect(requests, times=1).post(token_url, data=refresh_data1, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
            MockJsonResponse(200, {'access_token': access_token2, 'refresh_token': refresh_token2})
        )

        assert provider.get_token() == access_token2

    # 2nd refresh
    with freeze_time('2024-05-20 12:02:00'):
        refresh_data2 = {'client_id': AUTH_CLIENT_ID, 'grant_type': 'refresh_token', 'refresh_token': refresh_token2}
        access_token3 = make_token('Bearer', 300)
        refresh_token3 = make_token('Refresh', 3000)

        expect(requests, times=1).post(token_url, data=refresh_data2, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
            MockJsonResponse(200, {'access_token': access_token3, 'refresh_token': refresh_token3})
        )

        assert provider.get_token() == access_token3

    # logout
    with freeze_time('2024-05-20 12:03:00'):
        logout_data = {'client_id': AUTH_CLIENT_ID, 'refresh_token': refresh_token3}

        expect(requests, times=1).post(logout_url, data=logout_data, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
            MockJsonResponse(200, {})
        )

        provider.close()

    # new session
    with freeze_time('2024-05-20 12:04:00'):
        login_data2 = {
            'client_id': AUTH_CLIENT_ID,
            'grant_type': 'password',
            'username': auth_username,
            'password': auth_password,
        }
        access_token4 = make_token('Bearer', 300)
        refresh_token4 = make_token('Refresh', 3000)

        expect(requests, times=1).post(token_url, data=login_data2, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
            MockJsonResponse(200, {'access_token': access_token4, 'refresh_token': refresh_token4})
        )

        assert provider.get_token() == access_token4

    verifyNoUnwantedInteractions()
    unstub()


def test_token_client_login_fails_with_no_auth_server():
    """Tests that TokenClient raises ClientAuthenticationError if auth server has not been set"""

    provider = TokenClient('', '', '', '')
    provider._token_url = ''
    with pytest.raises(ClientConfigurationError, match='No auth server configured'):
        provider.get_token()

    verifyNoUnwantedInteractions()
    unstub()


def test_token_client_login_fails_with_wrong_password(auth_server_url, auth_realm, auth_username, auth_password):
    """Tests that TokenClient raises ClientAuthenticationError if server login fails"""

    token_url = f'{auth_server_url}/realms/{auth_realm}/protocol/openid-connect/token'
    login_data = {
        'client_id': AUTH_CLIENT_ID,
        'grant_type': 'password',
        'username': auth_username,
        'password': auth_password,
    }
    response_body = {'detail': 'invalid password'}

    # Prepare login
    expect(requests, times=1).post(token_url, data=login_data, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
        MockJsonResponse(401, response_body)
    )

    provider = TokenClient(auth_server_url, auth_realm, auth_username, auth_password)
    with pytest.raises(ClientAuthenticationError, match='Getting access token from auth server failed'):
        provider.get_token()

    verifyNoUnwantedInteractions()
    unstub()


def test_token_client_recovers_from_revoked_session(auth_server_url, auth_realm, auth_username, auth_password):
    """Tests that TokenClient starts a new session when getting new tokens using the refresh token fails."""

    refresh_token = make_token('Refresh', 3000)
    token_url = f'{auth_server_url}/realms/{auth_realm}/protocol/openid-connect/token'
    refresh_data = {
        'client_id': AUTH_CLIENT_ID,
        'grant_type': 'refresh_token',
        'refresh_token': refresh_token,
    }
    refresh_response_body = {'detail': 'session has been revoked'}

    # Prepare refresh request
    expect(requests, times=1).post(token_url, data=refresh_data, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
        MockJsonResponse(401, refresh_response_body)
    )

    login_data = {
        'client_id': AUTH_CLIENT_ID,
        'grant_type': 'password',
        'username': auth_username,
        'password': auth_password,
    }
    login_response_body = {
        'access_token': make_token('Bearer', 300),
        'refresh_token': make_token('Refresh', 3000),
    }

    # Prepare login request
    expect(requests, times=1).post(token_url, data=login_data, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
        MockJsonResponse(200, login_response_body)
    )

    provider = TokenClient(auth_server_url, auth_realm, auth_username, auth_password)
    provider._refresh_token = refresh_token

    assert provider.get_token() == login_response_body['access_token']

    verifyNoUnwantedInteractions()
    unstub()


def test_token_client_logout_fails(auth_server_url, auth_realm, auth_username, auth_password):
    """Tests that TokenClient raises ClientAuthenticationError if logout fails"""

    token_url = f'{auth_server_url}/realms/{auth_realm}/protocol/openid-connect/token'
    logout_url = f'{auth_server_url}/realms/{auth_realm}/protocol/openid-connect/logout'

    access_token1 = make_token('Bearer', 300)
    refresh_token1 = make_token('Refresh', 3000)

    login_data = {
        'client_id': AUTH_CLIENT_ID,
        'grant_type': 'password',
        'username': auth_username,
        'password': auth_password,
    }
    logout_data = {'client_id': AUTH_CLIENT_ID, 'refresh_token': refresh_token1}
    response_body = {'detail': 'unknown session'}

    # Prepare login
    expect(requests, times=1).post(token_url, data=login_data, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
        MockJsonResponse(200, {'access_token': access_token1, 'refresh_token': refresh_token1})
    )

    # Prepare logout
    expect(requests, times=1).post(logout_url, data=logout_data, timeout=AUTH_REQUESTS_TIMEOUT).thenReturn(
        MockJsonResponse(404, response_body)
    )

    provider = TokenClient(auth_server_url, auth_realm, auth_username, auth_password)
    assert provider.get_token() == access_token1
    with pytest.raises(ClientAuthenticationError, match='Logout failed'):
        provider.close()

    verifyNoUnwantedInteractions()
    unstub()


def _patch_env(patcher, **patched):
    for key in ['IQM_TOKEN', 'IQM_TOKENS_FILE', 'IQM_AUTH_SERVER', 'IQM_AUTH_USERNAME', 'IQM_AUTH_PASSWORD']:
        if patched.get(key):
            patcher(key, patched[key])
        else:
            patcher(key, '')


def test_token_manager_initialization_with_keyword_args(
    monkeypatch, auth_server_url, auth_username, auth_password
) -> None:
    """Test that TokenManager initializes correct token provider based on keyword arguments"""
    _patch_env(monkeypatch.setenv)

    token_manager = TokenManager()
    assert token_manager._token_provider is None

    token = make_token('Bearer', 300)
    token_manager = TokenManager(token=token)
    assert isinstance(token_manager._token_provider, ExternalToken)
    assert token_manager._token_provider._token == token

    path = '/some/path/to/tokens_file.json'
    token_manager = TokenManager(tokens_file=path)
    assert isinstance(token_manager._token_provider, TokensFileReader)
    assert token_manager._token_provider._path == path

    token_manager = TokenManager(auth_server_url=auth_server_url, username=auth_username, password=auth_password)
    assert isinstance(token_manager._token_provider, TokenClient)
    assert (
        token_manager._token_provider._token_url
        == f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/token'
    )
    assert (
        token_manager._token_provider._logout_url
        == f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout'
    )
    assert token_manager._token_provider._username == auth_username
    assert token_manager._token_provider._password == auth_password

    verifyNoUnwantedInteractions()
    unstub()


def test_token_manager_initialization_with_environment_vars(monkeypatch, auth_server_url, auth_username, auth_password):
    """Test that TokenManager initializes correct token provider based on environment variables"""
    _patch_env(monkeypatch.setenv)

    token_manager = TokenManager()
    assert token_manager._token_provider is None

    token = make_token('Bearer', 300)
    _patch_env(monkeypatch.setenv, **{'IQM_TOKEN': token})
    token_manager = TokenManager()
    assert isinstance(token_manager._token_provider, ExternalToken)
    assert token_manager._token_provider._token == token

    path = '/some/path/to/tokens_file.json'
    _patch_env(monkeypatch.setenv, **{'IQM_TOKENS_FILE': path})
    token_manager = TokenManager()
    assert isinstance(token_manager._token_provider, TokensFileReader)
    assert token_manager._token_provider._path == path

    _patch_env(
        monkeypatch.setenv,
        **{'IQM_AUTH_SERVER': auth_server_url, 'IQM_AUTH_USERNAME': auth_username, 'IQM_AUTH_PASSWORD': auth_password},
    )
    token_manager = TokenManager()
    assert isinstance(token_manager._token_provider, TokenClient)
    assert (
        token_manager._token_provider._token_url
        == f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/token'
    )
    assert (
        token_manager._token_provider._logout_url
        == f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout'
    )
    assert token_manager._token_provider._username == auth_username
    assert token_manager._token_provider._password == auth_password

    verifyNoUnwantedInteractions()
    unstub()


@pytest.mark.parametrize(
    'args,env',
    [
        # Token and some other parameter
        ({'token': 'some-token', 'tokens_file': 'some-path'}, {}),
        # Tokens file and some other parameter
        ({'tokens_file': 'some-path', 'username': 'some-user'}, {}),
        # Some auth server parameter is missing
        ({'auth_server_url': 'some-url', 'password': 'very-secret'}, {}),
        # Token and some other parameter as environment variables
        ({}, {'IQM_TOKEN': 'some-token', 'IQM_AUTH_USERNAME': 'some-path'}),
        # Tokens file and some other parameter
        ({}, {'IQM_TOKENS_FILE': 'some-path', 'IQM_AUTH_PASSWORD': 'very-secret'}),
        # Some auth server parameter is missing
        ({}, {'IQM_AUTH_USERNAME': 'some_user', 'IQM_AUTH_PASSWORD': 'very-secret'}),
    ],
)
def test_token_manager_invalid_combination_of_parameters(args, env, monkeypatch):
    """Test that TokenManager raises ClientConfigurationError if the parameters are ambiguous"""
    _patch_env(monkeypatch.setenv, **env)
    with pytest.raises(ClientConfigurationError, match='Invalid combination of authentication parameters specified'):
        TokenManager(**args)

    verifyNoUnwantedInteractions()
    unstub()


@pytest.mark.parametrize(
    'args,env',
    [
        # Mixed initialisation parameters and environment variables
        ({'tokens_file': 'some-path'}, {'IQM_TOKEN': 'some-token'}),
        # Mixed initialisation parameters and environment variables
        ({'auth_server_url': 'some-url'}, {'IQM_AUTH_USERNAME': 'some-user', 'IQM_AUTH_PASSWORD': 'very-secret'}),
    ],
)
def test_token_manager_mixed_source_of_parameters(args, env, monkeypatch):
    """Test that TokenManager raises ClientConfigurationError if the parameters are ambiguous"""
    _patch_env(monkeypatch.setenv, **env)
    error_message = 'Authentication parameters given both as initialisation args and as environment variables'
    with pytest.raises(ClientConfigurationError, match=error_message):
        TokenManager(**args)

    verifyNoUnwantedInteractions()
    unstub()


def test_token_manager_provides_bearer_token(monkeypatch):
    """Test that TokenManager provides bearer token"""

    _patch_env(monkeypatch.setenv)
    expected_token = make_token('Bearer', 300)
    mock_provider = mock(TokenProviderInterface)
    when(mock_provider).get_token().thenReturn(expected_token)

    token_manager = TokenManager()

    # When authentication is not configured get_bearer_token returns None
    assert token_manager._token_provider is None
    assert token_manager.get_bearer_token() is None

    # An existing valid access token is returned instead of asking token_provider for a new one
    token_manager._token_provider = mock_provider
    existing_token = make_token('Bearer', 300)
    token_manager._access_token = existing_token
    assert token_manager.get_bearer_token() == f'Bearer {existing_token}'

    # Otherwise get_bearer_token returns the token from the token_provider    token_manager._access_token = None
    token_manager._access_token = None
    assert token_manager.get_bearer_token() == f'Bearer {expected_token}'

    verifyNoUnwantedInteractions()
    unstub()


def test_token_manager_close(monkeypatch):
    """Test that TokenManager closes the token provider"""

    _patch_env(monkeypatch.setenv)
    mock_provider = mock(TokenProviderInterface)
    expect(mock_provider, times=1).close().thenReturn(None)

    # When authentication is not configured there is nothing to close
    token_manager = TokenManager()
    assert not token_manager.close()

    # TokenManager calls `close()` of the token provider, sets token provider to None and returns True
    token_manager._token_provider = mock_provider
    assert token_manager.close()
    assert token_manager._token_provider is None

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_gets_token(
    monkeypatch, base_url, dynamic_architecture_url, jobs_url, sample_dynamic_architecture, sample_circuit
):
    """Test that submit_circuits gets bearer token from TokenManager"""
    _patch_env(monkeypatch.setenv)
    token, client = _make_client_with_token(base_url)

    when(requests).get(
        dynamic_architecture_url,
        headers={'User-Agent': client._signature, 'Authorization': f'Bearer {token}'},
        timeout=REQUESTS_TIMEOUT,
    ).thenReturn(MockJsonResponse(200, json_data=sample_dynamic_architecture.model_dump()))

    expect(requests, times=1).post(
        jobs_url,
        json=ANY,
        headers={'User-Agent': client._signature, 'Authorization': f'Bearer {token}', 'Expect': '100-Continue'},
        timeout=REQUESTS_TIMEOUT,
    ).thenReturn(MockJsonResponse(200, json_data={'id': str(uuid4())}))

    assert isinstance(client.submit_circuits(circuits=[sample_circuit], shots=10), UUID)

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_gets_token(monkeypatch, base_url, jobs_url, ready_job_result):
    """Test that get_run gets bearer token from TokenManager"""
    _patch_env(monkeypatch.setenv)
    token, client = _make_client_with_token(base_url)
    job_id = uuid4()

    expect(requests, times=1).get(
        f'{jobs_url}/{str(job_id)}',
        headers={'User-Agent': client._signature, 'Authorization': f'Bearer {token}'},
        timeout=REQUESTS_TIMEOUT,
    ).thenReturn(ready_job_result)

    assert isinstance(client.get_run(job_id), RunResult)

    verifyNoUnwantedInteractions()
    unstub()


def test_get_run_status_gets_token(monkeypatch, base_url, jobs_url, pending_compilation_status):
    """Test that get_run gets bearer token from TokenManager"""
    _patch_env(monkeypatch.setenv)
    token, client = _make_client_with_token(base_url)
    job_id = uuid4()

    expect(requests, times=1).get(
        f'{jobs_url}/{str(job_id)}/status',
        headers={'User-Agent': client._signature, 'Authorization': f'Bearer {token}'},
        timeout=REQUESTS_TIMEOUT,
    ).thenReturn(pending_compilation_status)

    assert isinstance(client.get_run_status(job_id), RunStatus)

    verifyNoUnwantedInteractions()
    unstub()


def test_abort_job_gets_token(monkeypatch, base_url, jobs_url):
    """Test that abort_job gets bearer token from TokenManager"""
    _patch_env(monkeypatch.setenv)
    token, client = _make_client_with_token(base_url)
    job_id = uuid4()

    expect(requests, times=1).post(
        f'{jobs_url}/{str(job_id)}/abort',
        headers={'User-Agent': client._signature, 'Authorization': f'Bearer {token}'},
        timeout=REQUESTS_TIMEOUT,
    ).thenReturn(MockJsonResponse(200, {}))

    client.abort_job(job_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_close_auth_session(monkeypatch, base_url):
    """Test that closing auth session closes TokenManager"""
    _patch_env(monkeypatch.setenv)
    _, client = _make_client_with_token(base_url)
    client._token_manager = mock(TokenManager)
    expect(client._token_manager, times=1).close().thenReturn(True)

    assert client.close_auth_session()

    verifyNoUnwantedInteractions()
    unstub()


def test_close_auth_session_when_client_destroyed(monkeypatch, base_url):
    """Test that deleting client closes TokenManager"""
    _patch_env(monkeypatch.setenv)
    _, client = _make_client_with_token(base_url)
    client._token_manager = mock(TokenManager)
    expect(client._token_manager, times=1).close().thenReturn(True)

    del client

    verifyNoUnwantedInteractions()
    unstub()



# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the IQM client models."""
from uuid import UUID

import pytest

from iqm.iqm_client.models import (
    CircuitBatch,
    DynamicQuantumArchitecture,
    GateImplementationInfo,
    GateInfo,
    HeraldingMode,
    JobParameters,
    Metadata,
    RunRequest,
)


@pytest.fixture
def shots():
    return 1024


@pytest.fixture
def circuits_batch():
    return CircuitBatch()


@pytest.fixture
def heralding_mode():
    return HeraldingMode.ZEROS


@pytest.mark.parametrize(
    'metadata_factory',
    [
        # V1 and RESONANCE_V1
        lambda shots, circuits_batch, heralding_mode: Metadata(
            request=RunRequest(circuits=circuits_batch, shots=shots, heralding_mode=heralding_mode)
        ),
        # V2
        lambda shots, circuits_batch, heralding_mode: Metadata(
            parameters=JobParameters(shots=shots, heralding_mode=heralding_mode), circuits_batch=circuits_batch
        ),
    ],
)
def test_metadata(metadata_factory, shots, circuits_batch, heralding_mode):
    """Tests different modes of Metadata class initialization."""
    metadata = metadata_factory(shots, circuits_batch, heralding_mode)
    assert metadata.shots == shots
    assert metadata.circuits == circuits_batch
    assert metadata.heralding_mode == heralding_mode


def test_gate_info_loci():
    gate_info = GateInfo(
        implementations={
            'tgss': GateImplementationInfo(loci=(('QB1', 'QB2'), ('QB2', 'QB3'), ('QB10', 'QB2'), ('QB3', 'COMPR2'))),
            'crf': GateImplementationInfo(loci=(('QB3', 'QB1'), ('QB2', 'QB3'), ('QB2', 'COMPR1'), ('QB3', 'COMPR1'))),
        },
        default_implementation='tgss',
        override_default_implementation={},
    )
    assert gate_info.loci == (
        ('QB1', 'QB2'),
        ('QB2', 'COMPR1'),
        ('QB2', 'QB3'),
        ('QB3', 'COMPR1'),
        ('QB3', 'COMPR2'),
        ('QB3', 'QB1'),
        ('QB10', 'QB2'),
    )


def test_dqa_components():
    dqa = DynamicQuantumArchitecture(
        calibration_set_id=UUID('59478539-dcef-4b2e-80c8-122d7ec3fc89'),
        qubits=['QB3', 'QB1', 'QB11', 'QB20', 'QB2', 'QB10'],
        computational_resonators=['COMPR2', 'COMPR1'],
        gates={},
    )
    assert dqa.components == ('COMPR1', 'COMPR2', 'QB1', 'QB2', 'QB3', 'QB10', 'QB11', 'QB20')
